static const char *copyright5 = "copyright (c) 2011 Victor Milenkovic and Elisha Sacks.  Permission required for commercial use.  Contact vjm@cs.miami.edu";

#include "math.h"

class PV {
public:
  double x, y;
  PV () : x(0), y(0) {}
  PV (double x, double y) : x(x), y(y) {}
  PV (double q) : x(cos(q)), y(sin(q)) {}

  PV rot90 () { return PV(-y, x); }
  PV operator* (double s) { return PV(x*s, y*s); }
  PV operator/ (double s) { return PV(x/s, y/s); }
  double operator* (PV that) { return x * that.x + y * that.y; }
  double operator% (PV that) { return x * that.y - y * that.x; }
  double length () { return sqrt(*this * *this); }
  PV unit () { return *this / length(); }
  PV conjugate () { return PV(x, -y); }
  PV times (PV that) { return PV(x * that.x - y * that.y,
                                 x * that.y + y * that.x); }
  PV operator- (PV that) { return PV(x - that.x, y - that.y); }
  PV operator+ (PV that) { return PV(x + that.x, y + that.y); }
  double distance (PV that) { return (*this - that).length(); }
  PV operator- () { return PV(-x, -y); }
  double angle () { return atan2(y, x); }
  PV nthRoot (int n) {
    double q = atan2(y, x) / n;
    return PV(cos(q), sin(q));
  }
  bool operator== (const PV &b) const { return x == b.x && y == b.y; }

  class Compare {
  public:
    bool operator() (const PV &a, const PV &b) {
      if (a.x != b.x)
        return a.x < b.x;
      return a.y < b.y;
    }
  };
};

class PCR {
public:
  PV p, c;
  double r;

  PCR () : r(0) {}
  PCR (PV p, PV c, double r) : p(p), c(c), r(r) {}
  PCR operator* (double s) { return PCR(p * s, c * s, r * s); }
};

