#include "acp.h"
#include "PolyArc.h"
#include <typeinfo>
#include <map>
#include <iomanip>
using namespace std;

// Intersect circle (0,0),r1 with c,r2
bool intersectCircles3 (const Parameter &r1, Point2 c, const Parameter &r2,
                        Point2 q[2], Point2 cq[2]) {
  static int count;
  if (++count == 0)
    cout << "intersectCircles this is it" << endl;

  // assert((r1 - r2).sign() <= 0);

  Parameter d = c.norm();
  Parameter x;

  if (r1.round() < d.round()) {
    Parameter e = d - r2;

    if ((r1 + e).sign() < 0 || (r1 - e).sign() < 0)
      return false;
    
    Parameter r2Me2D2d = (r1 * r1 - e * e) / d / 2;
    x = e + r2Me2D2d;
  }
  else {
    Parameter f = r2 - r1;
    if ((d - f).sign() < 0)
      return false;

    Parameter d2Mf2D2d = (d * d - f * f) / d / 2;
    Parameter fDd = f / d;
    Parameter rfDd = r1 * fDd;
    x = d2Mf2D2d - rfDd;
  }

  Parameter r2Mx2 = r1 * r1 - x * x;
  Parameter y = r2Mx2.sqrt();
  assert(!isnan(y.round()));

  Point2 u = c / d;
  Point2 v = u.rot90();
  Point2 p = u * x;
  q[0] = p + v * y;
  q[1] = p - v * y;
  cq[0] = q[0] - c;
  cq[1] = q[1] - c;

  static double maxRR;
  for (int i = 0; i < 2; i++) {
    PV qi(P2PV(q[i]));
    assert(!isnan(qi.x) && !isnan(qi.y));
    double rr = fabs(qi.length() - r1.round());
    if (maxRR < rr) {
      cout << "maxRR1 " << rr << endl;
      maxRR = rr;
    }

    PV cqi(P2PV(cq[i]));
    assert(!isnan(cqi.x) && !isnan(cqi.y));
    rr = fabs(cqi.length() - r2.round());
    if (maxRR < rr) {
      cout << "maxRR2 " << rr << endl;
      maxRR = rr;
    }
  }

  return true;
}

bool intersectCircles (const Parameter &r1, Point2 c, const Parameter &r2,
                       Point2 q[2], Point2 cq[2]) {
  static int count;
  if (++count == 0)
    cout << "intersectCircles this is it" << endl;

  Parameter d = c.norm();

  if (r1.round() < r2.round()) // (r1.weakCompare(r2) <= 0)
    return intersectCircles3(r1, c, r2, q, cq);

  Point2 cm = -c;
  if (!intersectCircles3(r2, cm, r1, cq, q))
    return false;

  Point2 temp;
  temp = q[0];
  q[0] = q[1];
  q[1] = temp;
  temp = cq[0];
  cq[0] = cq[1];
  cq[1] = temp;

  /*
  Point2 q2[2];
  assert(intersectCircles1(r1, c, r2, q2));
  assert((q[0] - q[1]).innerproductSign(q2[0] - q2[1]) > 0);
  */

  return true;
}

void CircleTHS::calculate () {
  double p2 = 2;
  // if (t->getP().x.l == 1e20) p2.increasePrecision();
  Point2 m = (t->getP() + h->getP()) / p2;
  Point2 th = h->getP() - t->getP();
  Point2 n = th.unit().rot90();
  Parameter ss = s->getX();
  Parameter l = (ss * ss - (th.innerproduct(th)) / (p2*p2)).sqrt();
  c = ss.sign() > 0 ? m + n * l : m - n * l;
}

int IntCC::construct (Circle *a, Circle *b, IntCC *out[2]) {
  const char *aType = typeid(*a).name();
  const char *bType = typeid(*b).name();
  try {
    int i = (!(typeid(*a) == typeid(CirclePS) &&
               typeid(*b) == typeid(CirclePS) &&
               dynamic_cast<CirclePS*>(a)->s ==
               dynamic_cast<CirclePS*>(b)->s) &&
             b->getS().abs() < a->getS().abs());
    assert(i == 0 || i == 1);
    if (i) {
      Circle *tmp = a;
      a = b;
      b = tmp;
    }
    Point2 oq[2], cq[2];
    
    if (!intersectCircles(a->getS().abs(), b->getC() - a->getC(),
                          b->getS().abs(), oq, cq))
      return 0;
    out[0] = new IntCC(oq[0], 0, a, b);
    out[1] = new IntCC(oq[1], 1, b, a);
    return 2;
  } catch (SignException e) {
    a->increasePrecision();
    b->increasePrecision();
    int n = construct(a, b, out);
    if (n == 2) {
      out[0]->decreasePrecision();
      out[1]->decreasePrecision();
    }
    a->decreasePrecision();
    b->decreasePrecision();
    return n;
  }
}

void IntCC::increasePrecision () {
  if (cp.x.l == 1e20)
    return;

  static int count;
  if (++count == 0)
    cout << "this is it increasePrecision" << endl;

  c[0]->increasePrecision();
  c[1]->increasePrecision();

  int i = iSmaller;
  Circle *a = c[i];
  Circle *b = c[!i];
  Point2 oq[2], cq[2];
  bool check = intersectCircles(a->getS().abs(), b->getC() - a->getC(),
                                b->getS().abs(), oq, cq);
  assert(check);
  cp = oq[i];
}  

void IntCC::decreasePrecision () {
  if (cp.x.l != 1e20)
    return;
  c[0]->decreasePrecision();
  c[1]->decreasePrecision();
  cp.decreasePrecision();
}

void PolyArc::Edge::increasePrecision () {
  c->increasePrecision();
  t->increasePrecision();
  h->increasePrecision();
}

void PolyArc::Edge::decreasePrecision () {
  c->decreasePrecision();
  t->decreasePrecision();
  h->decreasePrecision();
}

PolyArc::Edge::Edge (PolyArc *poly, Vert *t, Vert *h)
  : poly(poly), id(poly->edges.size()), t(t), h(h), c(0) {
  poly->edges.push_back(this);
  t->t = this;
  h->h = this;
  concave = t->n->getN().crossproduct(h->n->getN()) < 0;
}

PolyArc::Edge::Edge (PolyArc *poly, Vert *t, Vert *h, bool concave)
  : poly(poly), id(poly->edges.size()), t(t), h(h), c(0) {
  poly->edges.push_back(this);
  t->t = this;
  h->h = this;
  this->concave = concave;
}

PolyArc::Edge::Edge (PolyArc *poly, Vert *tv, Vert *hv,
                     Point *tp, Point *hp, Param *s)
  : poly(poly), id(poly->edges.size()), t(tv), h(hv), c(0) {
  poly->edges.push_back(this);
  c = new CircleTHS(tp, hp, s);
  t->t = this;
  t->n = new NormalCP(c, tp);
  h->h = this;
  h->n = new NormalCP(c, hp);
  concave = s < 0;
}

PolyArc::Edge::Edge (PolyArc *poly, Vert *t, Vert *h, Circle *c)
  : poly(poly), id(poly->edges.size()), t(t), h(h), c(c) {
  poly->edges.push_back(this);
  t->t = this;
  h->h = this;
  concave = c->getS() < 0;
}

PolyArc::Edge::Edge (PolyArc *poly, Vert *t, Vert *h, Edge *e)
  : poly(poly), id(poly->edges.size()), t(t), h(h), c(e->c) {
  poly->edges.push_back(this);
  t->t = this;
  h->h = this;
  concave = e->concave;
}

bool PolyArc::Edge::contains (IntCC *intCC) {
  return compare(t, intCC) < 0 && compare(h, intCC) > 0;
}

int PolyArc::Edge::compare (Vert *v, IntCC *intCC) {
  try {
    assert(v == t || v == h);
    Point2 vn = v->n->getN();
    Point2 in = intCC->getN(c);
    int sign = -vn.crossproduct(in).sign();
    return concave ? -sign : sign;
  } catch (SignException e) {
    v->n->increasePrecision();
    intCC->increasePrecision();
    int ret = compare(v, intCC);
    v->n->decreasePrecision();
    intCC->decreasePrecision();
    return ret;
  }
}

int PolyArc::Edge::compare (Intn *a, Intn *b) {
  Point2 an = a->intCC->getN(c);
  Point2 bn = b->intCC->getN(c);
  int sign = -an.crossproduct(bn).sign();
  return concave ? -sign : sign;
}

bool PolyArc::Edge::closer (Point *p, Param *d) {
  if (c == 0)
    return t->n->getP().distance(p->getP()) < d->getX();

  Point2 cp = p->getP() - c->getC();

  if (t->n->getN().crossproduct(cp).sign() < 0 ||
      h->n->getN().crossproduct(cp).sign() > 0)
    return false;

  Parameter cpLen = cp.norm();
  Parameter r = c->getS().abs();

  /*
  if (cpLen < r)
    // dist = r - cp
    // dist < d <--> r - cp < d <--> d + cp - r > 0
    return r - cpLen < d;
  else
    // dist = cp - r
    // dist < d <--> cp - r < d <--> d + r - cp > 0
    return cpLen - r < d;
  */

  return (cpLen - r).abs() < d->getX();
}

int PolyArc::Intn::construct (Edge *a, Edge *b, Intn *intns[2]) {
  if (a->t != 0 && a->t == b->h ||
      a->h != 0 && a->h == b->t)
    return 0;

  IntCC *intCCs[2];
  int nIntCC = IntCC::construct(a->c, b->c, intCCs);
  if (nIntCC == 0)
    return 0;

  int nIntn = 0;
  for (int i = 0; i < nIntCC; i++) {
    IntCC *intCC = intCCs[i];
    if (!(a->contains(intCC) && b->contains(intCC))) {
      delete intCC;
      continue;
    }
    if ((a->concave == b->concave) == (a->c == intCC->c[0]))
      intns[nIntn++] = new Intn(intCC, a, b);
    else
      intns[nIntn++] = new Intn(intCC, b, a);
  }
  return nIntn;
}

void PolyArc::Intn::insert (Edge *e, vector<Intn *> &intns) {
  int a = 0, b = intns.size() - 1;

  while (a <= b) {
    int c = (a + b) / 2;
    if (e->compare(this, intns[c]) < 0)
      b = c - 1;
    else
      a = c + 1;
  }
  intns.insert(intns.begin() + a, this);
}

void PolyArc::init (vector<PV> &pointsI, vector<double> radiiI,
                    vector<int> &tails, vector<int> &heads) {
  int n = pointsI.size();
  assert(n == radiiI.size());
  assert(n == tails.size());
  assert(n == heads.size());

  points.resize(n);
  for (int i = 0; i < n; i++)
    points[i] = new PointXY(Point2(Parameter::input(pointsI[i].x),
                                   Parameter::input(pointsI[i].y)));

  radii.resize(n);
  for (int i = 0; i < n; i++)
    radii[i] = new ParamX(Parameter::input(radiiI[i]));

  // Duplicate all vertices to leave room for smoothing arcs.
  for (int i = 0; i < n; i++) {
    Vert *u = new Vert(this);
    Vert *v = new Vert(this);
  }

  // Create regular arcs.
  for (int i = 0; i < n; i++)
    Edge *e = new Edge(this, verts[2*tails[i]+1], verts[2*heads[i]],
                       points[tails[i]], points[heads[i]], radii[i]);

  // Create smoothing arcs.
  for (int i = 0; i < n; i++)
    Edge *e = new Edge(this, verts[2*i], verts[2*i+1]);
}

PolyArc::PolyArc (const char *fileName, double angle)
  : parent(0), o(0) {
  PV rot(cos(angle), sin(angle));

  ifstream in;
  in.open(fileName);
  if (!in) {
    cerr << "PolyArc::PolyArc:  Could not open " << fileName << endl;
    return;
  }

  int nVerts;
  in >> nVerts;
  vector<PV> pointsI(nVerts);
  for (int i = 0; i < nVerts; i++) {
    in >> pointsI[i].x >> pointsI[i].y;
    pointsI[i] = rot.times(pointsI[i]);
  }

  int nEdges;
  in >> nEdges;
  assert(nEdges == nVerts);
  vector<double> radiiI(nEdges);
  vector<int> tails(nEdges);
  vector<int> heads(nEdges);
  for (int i = 0; i < nEdges; i++) {
    in >> radiiI[i] >> tails[i] >> heads[i];
#ifdef PERTURB
    if (i == 4 || i == 9)
      continue;
    radiiI[i] *= 1 + random(-0.01, 0.01);
    while (fabs(radiiI[i]) <= pointsI[tails[i]].distance(pointsI[heads[i]]) / 2)
      radiiI[i] *= 1 + random(0, 0.5);
#endif      
  }

  in.close();

  init(pointsI, radiiI, tails, heads);
}

void blah () {
  cout << "this is it" << endl;
}

PolyArc::PolyArc (PolyArc *poly, double oI)
  : parent(poly), o(new ParamX(Parameter::input(oI))) {
  Parameter x = o->getX();
  int n = parent->verts.size();
  assert(parent->edges.size() == n);

  for (int i = 0; i < n; i++) {
    Vert *vP = parent->verts[i];
    new Vert(this, new NormalNO(vP->n, o));
  }

  for (int i = 0; i < n; i++) {
    if (i == 116)
      blah();
    Edge *eP = parent->edges[i];
    Vert *t = verts[eP->t->id];
    Vert *h = verts[eP->h->id];
    Circle *c = 0;
    if (eP->c == 0 && !eP->concave)
      c = new CirclePS(eP->t->n, o);
    else if (eP->c != 0 && (!eP->concave || o->getX() < eP->c->getS().abs()))
      c = new CircleCO(eP->c, o);
    else
      continue;
    new Edge(this, t, h, c);
  }
}

PolyArc *PolyArc::levelSet () {
  assert(o != 0 && parent != 0);
  vector<vector<Intn *> > intnss(edges.size());

  for (int i = 0; i < edges.size(); i++) {
    for (int j = i+1; j < edges.size(); j++) {
      Intn *intns[2];
      int nIntns = Intn::construct(edges[i], edges[j], intns);
      for (int k = 0; k < nIntns; k++) {
        intns[k]->insert(edges[i], intnss[i]);
        intns[k]->insert(edges[j], intnss[j]);
      }
    }
  }
  
  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    vector<Intn *> &intns = intnss[i];
    for (int j = 0; j < intns.size(); j++) {
      Intn *intn = intns[j];
      if (e == intn->e[0])
        intn->ids[0] = j;
      else {
        assert(e == intn->e[1]);
        intn->ids[1] = j;
      }
    }
  }

  vector<vector<char> > danglings(edges.size());

  for (int i = 0; i < edges.size(); i++)
    danglings[i].resize(intnss[i].size()+1, false);

  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    vector<Intn *> &intns = intnss[i];
    for (int j = 0; j <= intns.size(); j++) {
      SubE sub(e, j);
      if (sub.next(intnss).isNull()) {
        SubE curr = sub;
        do {
          danglings[curr.e->id][curr.i] = true;
          curr = curr.prev(intnss);
        } while (!curr.isNull());
      }
      if (sub.prev(intnss).isNull()) {
        SubE curr = sub;
        do {
          danglings[curr.e->id][curr.i] = true;
          curr = curr.next(intnss);
        } while (!curr.isNull());
      }
    }
  }

  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    vector<Intn *> &intns = intnss[i];
    vector<char> &dangling = danglings[i];
    for (int j = 0; j <= intns.size(); j++) {
      if (dangling[j])
        continue;
      Point *p = 0;
      if (j < intns.size())
        p = intns[j]->intCC;
      else
        p = e->h->n;
      SubE sub(e, j);
      SubE subNext = sub.next(intnss);
      if (!parent->insideD(p, o))
        continue;
      SubE curr = sub;
      do {
        danglings[curr.e->id][curr.i] = true;
        curr = curr.next(intnss);
      } while (curr != sub);
    }
  }

  PolyArc *lev = new PolyArc(this);

  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    vector<Intn *> &intns = intnss[i];
    vector<char> &dangling = danglings[i];
    for (int j = 0; j <= intns.size(); j++) {
      if (dangling[j])
        continue;
      SubE first(e, j);
      SubE curr = first;
      Vert *firstVert = 0, *prevVert = 0;
      Edge *firstEdge = 0;
      do {
        Vert *v1 = 0, *v2 = 0;
        if (curr.i == intnss[curr.e->id].size())
          v1 = new Vert(lev, curr.e->h->n);
        else {
          Intn *intn = intnss[curr.e->id][curr.i];
          assert(curr.e == intn->e[1]);
          v1 = new Vert(lev, intn->intCC, 1);
          v2 = new Vert(lev, intn->intCC, 0);
          intn->intCC = 0;
        }
        if (firstVert == 0)
          firstVert = v1;
        else
          new Edge(lev, prevVert, v1, curr.e);
        if (v2 == 0)
          prevVert = v1;
        else {
          new Edge(lev, v1, v2, true);
          prevVert = v2;
        }
        assert(!danglings[curr.e->id][curr.i]);
        danglings[curr.e->id][curr.i] = true;
        curr = curr.next(intnss);
      } while (curr != first);
      new Edge(lev, prevVert, firstVert, first.e);
    }
  }

  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    vector<Intn *> &intns = intnss[i];
    for (int j = 0; j < intns.size(); j++) {
      Intn *intn = intns[j];
      if (intn->e[0] != 0)
        intn->e[0] = 0;
      else
        delete intns[j];
    }
  }

  return lev;
}

bool PolyArc::insideD (Point *p, Param *o) {
  GeoObj *ignore[2] = { 0, 0 };

  if (typeid(*p) == typeid(NormalNO)) {
    // p is original convolution vertex
    ignore[0] = dynamic_cast<NormalNO*>(p)->pn;
    // pn is normal to generating polyarc vertex
    // e->t->n == ignore[0] || e->h->n == ignore[0]
  }
  else if (typeid(*p) == typeid(IntCC)) {
    // p is intersection of two convolution arcs
    IntCC *intCC = dynamic_cast<IntCC*>(p);
    // intCC is intersection of two convolution circles
    for (int i = 0; i < 2; i++) {
      Circle *c = intCC->c[i];
      if (typeid(*c) == typeid(CircleCO)) {
        // circle is offset of polyarc arc
        ignore[i] = dynamic_cast<CircleCO*>(c)->circle;
        // e->c != 0 && e->c == ignore[i]
      }
      else if (typeid(*c) == typeid(CirclePS)) {
        // circle is offset of smoothing arc
        ignore[i] = dynamic_cast<CirclePS*>(c)->c;
        // e->c == 0 && e->t->n == ignore[i]
      }
      else
        assert(0);
    }
  }
  else
    assert(0);

  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    if (ignore[0] != 0 && (e->t->n == ignore[0] || e->h->n == ignore[0]) ||
        e->c != 0 && (e->c == ignore[0] || e->c == ignore[1]) ||
        e->c == 0 && (e->t->n == ignore[0] || e->t->n == ignore[1]))
      continue;
    // if (e->closer(p, o))
    // return true;
  }

  return false;
}

PolyArc::SubE PolyArc::SubE::next (vector<vector<PolyArc::Intn *> > &intnss) {
  if (i == intnss[e->id].size()) {
    if (e->h->t == 0)
      return SubE(0, -1);
    else
      return SubE(e->h->t, 0);
  }
  Intn *intn = intnss[e->id][i];
  if (e == intn->e[0])
    return SubE(0, -1);
  return SubE(intn->e[0], intn->ids[0]+1);
}

PolyArc::SubE PolyArc::SubE::prev (vector<vector<PolyArc::Intn *> > &intnss) {
  if (i == 0) {
    if (e->t->h == 0)
      return SubE(0, -1);
    else
      return SubE(e->t->h, intnss[e->t->h->id].size());
  }
  Intn *intn = intnss[e->id][i-1];
  if (e == intn->e[1])
    return SubE(0, -1);
  return SubE(intn->e[1], intn->ids[1]);
}

void PolyArc::drawFig (const char *fileName) {
  ofstream out;
  out.open(fileName);
  if (!out) {
    cerr << "PolyArc::drawFig:  Could not open " << fileName << endl;
    return;
  }

  out << "#FIG 3.2  Produced by xfig version 3.2.5" << endl
      << "Landscape" << endl
      << "Center" << endl
      << "Inches" << endl
      << "Letter  " << endl
      << "100.00" << endl
      << "Single" << endl
      << "-2" << endl
      << "1200 2" << endl;

  // PV shift(1 - sin(M_PI/5) / sin(M_PI/4), 0);
  PV shift(0, 0);
  cout << setprecision(23);
  cout << "shiftx " << shift.x << endl;
  cout << (2 * M_PI / 4) << endl;

  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    if (e->c == 0)
      continue;
    double r = e->c->getS().round();
    PV t(P2PV(e->t->n->getP()) + shift);
    PV h(P2PV(e->h->n->getP()) + shift);
    PV n = (t - h).rot90().unit();
    PV c(P2PV(e->c->getC()) + shift);
    PV m = c + n * r;

    static double minTH = 1e9;
    double distTH = t.distance(h);
    if (distTH < minTH) {
      cout << "minTH " << distTH << endl;
      minTH = distTH;
    }

    if (r < 0) {
      PV temp = t;
      t = h;
      h = temp;
    }

    if (t.distance(h) > 0.01)
      out << "5 1 0 1 0 7 " << i << " -1 -1 0.000 0 1 0 0"
          << " " << (1200*c.x) << " " << (-1200*c.y)
          << " " << (int) rint(1200*t.x) << " " << (int) rint(-1200*t.y)
          << " " << (int) rint(1200*m.x) << " " << (int) rint(-1200*m.y)
          << " " << (int) rint(1200*h.x) << " " << (int) rint(-1200*h.y)
          << endl;
    else
      out << "2 1 0 1 0 7 50 -1 -1 0.000 0 0 -1 0 0 2" << endl
          << "\t"
          << " " << (int) rint(1200*t.x) << " " << (int) rint(-1200*t.y)
          << " " << (int) rint(1200*h.x) << " " << (int) rint(-1200*h.y)
          << endl;
  }

  out.close();
}

void PolyArc::write (const char *fileName) {
  ofstream out;
  out.open(fileName);
  if (!out) {
    cerr << "PolyArc::write:  Could not open " << fileName << endl;
    return;
  }

  out << verts.size() << endl;
  for (int i = 0; i < verts.size(); i++) {
    Point2 p = verts[i]->n->getP();
    out << p.x.round() << " " << p.y.round() << endl;
  }
  out << endl;

  int n = 0;
  for (int i = 0; i < edges.size(); i++)
    if (edges[i]->c != 0)
      n++;

  out << n << endl;
  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    if (e->c != 0)
      out << e->c->getS().round() << " " << e->t->id << " " << e->h->id << endl;
  }

  out.close();
}

void PolyArc::write (const char *fileName, double sOff) {
  ofstream out;
  out.open(fileName);
  if (!out) {
    cerr << "PolyArc::write:  Could not open " << fileName << endl;
    return;
  }
  out << setprecision(23);

  vector<PV> vertVec;
  map<PV, int, PV::Compare> vertSet;
  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    PV t(e->t->n->getP().x.round(), e->t->n->getP().y.round());
    PV h(e->h->n->getP().x.round(), e->h->n->getP().y.round());
    if (vertSet.find(t) == vertSet.end()) {
      vertSet[t] = vertVec.size();
      vertVec.push_back(t);
    }
    if (vertSet.find(h) == vertSet.end()) {
      vertSet[h] = vertVec.size();
      vertVec.push_back(h);
    }
  }

  out << vertVec.size() << endl;
  for (int i = 0; i < vertVec.size(); i++) {
    out << vertVec[i].x << " " << vertVec[i].y << endl;
  }
  out << endl;

  int n = 0;
  for (int i = 0; i < edges.size(); i++)
    if (edges[i]->c != 0)
      n++;

  out << n << endl;
  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    if (e->c != 0) {
      double s = e->c->getS().round();
      if (s == sOff)
        s *= 0.75;
      PV t(e->t->n->getP().x.round(), e->t->n->getP().y.round());
      PV h(e->h->n->getP().x.round(), e->h->n->getP().y.round());
      out << -s << " " << vertSet[h] << " " << vertSet[t] << endl;
    }
    else {
      PV tn(e->t->n->getN().x.round(), e->t->n->getN().y.round());
      PV hn(e->h->n->getN().x.round(), e->h->n->getN().y.round());
      cerr << i << " " << (tn%hn) << endl;
    }
  }

  out.close();
}
