#include "acp.h"

double randomNumber (double rmin, double rmax)
{
  return rmin + (rmax - rmin)*random()/double(RAND_MAX);
}

double getTime ()
{
  timeval tv;
  gettimeofday(&tv, 0);
  return tv.tv_sec + 1e-6*tv.tv_usec;
}

double Parameter::delta = 1e-8;
unsigned int Parameter::precision = 250u;
SignException signException;

MValue::MValue (double rl, double ru) : refCnt(1)
{
  mpfr_init2(l, Parameter::precision);
  mpfr_set_d(l, rl, GMP_RNDD);
  mpfr_init2(u, Parameter::precision);
  mpfr_set_d(u, ru, GMP_RNDU);
}

void MValue::decRef ()
{ 
  if (--refCnt == 0) {
    mpfr_clear(l);
    mpfr_clear(u);
    delete this; 
  }
}

bool Parameter::operator< (double b) const
{ 
  return (b - *this).sign() == 1; 
}

bool Parameter::operator> (double b) const 
{ 
  return b < *this; 
}

Parameter operator+ (double a, const Parameter &b)
{
  return b + a;
}

Parameter operator- (double a, const Parameter &b)
{
  return (- b) + a;
}

Parameter operator* (double a, const Parameter &b)
{
  return b*a;
}

bool operator< (double a, const Parameter &b)
{
  return (b - a).sign() == 1;
}

bool operator> (double a, const Parameter &b)
{
  return b < a;
}

// debug

double pv (mpfr_t a)
{
  return mpfr_get_d(a, GMP_RNDN);
}

double iw (Parameter &p)
{
  return p.intervalWidth();
}
