#include "crit.h"

int criticality3 (Point2 *m, Point2 *f, Parameter *r, Point2 *th)
{
  PPoly p = criticality3eq(m, f, r);
  return solveCS(p, th);
}

PPoly criticality3eq (Point2 *m, Point2 *f, Parameter *r)
{
  Parameter hef[3], hfg[3], hge[3], hefhef[6], hfghfg[6], hgehge[6],
    hefhfg[6], eq[10],
    rere = r[0]*r[0], rfrf = r[1]*r[1], rgrg = r[2]*r[2];
  distanceSquared(m[0], f[0], m[1], f[1], hef);
  hef[2] = hef[2] - rere - rfrf;
  distanceSquared(m[1], f[1], m[2], f[2], hfg);
  hfg[2] = hfg[2] - rfrf - rgrg;
  distanceSquared(m[0], f[0], m[2], f[2], hge);
  hge[2] = hge[2] - rere - rgrg;
  product11(hef, hef, hefhef);
  product11(hfg, hfg, hfghfg);
  product11(hge, hge, hgehge);
  product11(hef, hfg, hefhfg);
  product12(hge, hefhfg, eq);
  for (int i = 0; i < 6; ++i)
    eq[4+i] = eq[4+i] + rere*hfghfg[i] + rfrf*hgehge[i] + rgrg*hefhef[i];
  eq[9] = eq[9] - rere*rfrf*rgrg*4.0;
  return functionCS3(eq);
}

// (theta*(a-c) + b-d)^2
void distanceSquared (const Point2 &a, const Point2 &b, 
		      const Point2 &c, const Point2 &d, Parameter *x)
{
  Parameter ux = (a.getX() - c.getX()), uy = (a.getY() - c.getY()),
    vx = (b.getX() - d.getX()), vy = (b.getY() - d.getY()),
    k1 = ux*vy - uy*vx, k2 = ux*vx + uy*vy;
  x[0] = k1 + k1;
  x[1] = k2 + k2;
  x[2] = ux*ux + uy*uy + vx*vx + vy*vy;
}

//(k0*s + k1*c + k2)*(l0*s + l1*c + l2)
void product11 (Parameter *k, Parameter *l, Parameter *x)
{
  x[0] = k[0]*l[0];
  x[1] = k[0]*l[1] + k[1]*l[0];
  x[2] = k[1]*l[1];
  x[3] = k[0]*l[2] + k[2]*l[0];
  x[4] = k[1]*l[2] + k[2]*l[1];
  x[5] = k[2]*l[2];
}

//(k0*s + k1*c + k2)*(l0*s^2 + l1*sc + l2*c^2 + l3*s + l4*c + l5)
void product12 (Parameter *k, Parameter *l, Parameter *x)
{
  x[0] = k[0]*l[0];
  x[1] = k[0]*l[1] + k[1]*l[0];
  x[2] = k[0]*l[2] + k[1]*l[1];
  x[3] = k[1]*l[2];
  x[4] = k[0]*l[3] + k[2]*l[0];
  x[5] = k[0]*l[4] + k[1]*l[3] + k[2]*l[1];
  x[6] = k[1]*l[4] + k[2]*l[2];
  x[7] = k[0]*l[5] + k[2]*l[3];
  x[8] = k[1]*l[5] + k[2]*l[4];
  x[9] = k[2]*l[5];
}

PPoly functionCS3 (Parameter *a)
{
  int m[20] = {0, 3, 1, 2, 2, 1, 3, 0, 0, 2, 1, 1, 2, 0, 0, 1, 1, 0, 0, 0};
  int i = 0;
  while (a[i].lb() == 0.0 && a[i].ub() == 0.0)
    ++i;
  return PPoly(2, 10 - i, a + i, m + 2*i);
}

