#ifndef EIGEN
#define EIGEN

#include "acp.h"
//#include <evd.h>
//#include <densesolver.h>

int realRoots (const PFunPoly &f, double *res);

int realRootsR (const PFunPoly &f, double *res);

void companionMatrixR (const PFunPoly &f, double *a);

int eigenvaluesR (double *a, int k, double *res);

class Monomial {
 public:
  Monomial (int in = 0);
  Monomial (const Monomial &b);
  ~Monomial () { delete [] elts; }
  Monomial & operator= (const Monomial &b);
  int degree () const;
  bool operator< (const Monomial &b) const;
  bool inI0 (PPoly *f) const;

  int n, *elts;
};

class MonomialOrder {
 public:
  MonomialOrder (int ni, PPoly *fi) : n(ni), f(fi) {}
  bool operator() (const Monomial &a, const Monomial &b) const;

  int n;
  PPoly *f;
};

int systemDegree (int n, PPoly *f);

int maxMonomialDegree (int n, PPoly *f);

int formMonomials (int n, PPoly *f, int dmax, Monomial *&z);

int choose (int k, int n);

int Zindex (const Monomial &m, int N, Monomial *z, const MonomialOrder &mo);

int * varIndex (int n, PPoly *f, int N, Monomial *z);

int realRoots (int n, PPoly *f, double *res);

int realRootsR (int n, PPoly *f, double *res);

double * setupA1R (int n, PPoly *f, int m, int N, Monomial *z);

double * setupA0R (int n, PPoly *f, int m, int N, Monomial *z);

int eigenRootsR (int n, PPoly *f, int m, double *evals, double *evalsi, 
		 double *evecs, int nr, double *a10, int *ind, double *res);

void eigenRootR (int n, int m, double eval, double *evec, int nr, double *a10,
		 int *ind, double *res);

// lapack interface

bool linsolve (double *a, double *b, int m, int k, double maxcond);

double maxnorm (double *a, int m, int n);

bool eigen (int k, double *a, double *evalsr, double *evalsi, double *evecs, 
	    double maxcond);

#ifdef __APPLE__
#include <Accelerate/Accelerate.h>
#endif
extern "C" {
#ifndef __APPLE__
  #include <cblas.h>
#endif
  int dgetrf_ (int *m, int *n, double *a, int *lda, int *ipiv, int *info);
  int dgecon_ (char *norm, int *n, double *a, int *lda, double *anorm,
               double *rcond, double *work, int *iwork, int *info);
  int dgetrs_ (char *trans, int *n, int *nrhs, double *a, int *lda, int *ipiv,
               double *b, int *ldb, int *info);
  int dgeevx_ (char *balanc, char *jobvl, char *jobvr, char *sense, int *n,
               double *a, int *lda, double *wr, double *wi,
               double *vl, int *ldvl, double *vr, int *ldvr,
               int *ilo, int *ihi, double *scale, double *abnrm,
               double *rconde, double *rcondv, double *work,
               int *lwork, int *iwork, int *info);
}

#endif
