#ifdef HPOINT
#include "hpoint.h"
#else
#include "point.h"
#endif

Point2 operator* (const Parameter &k, const Point2 &a)
{
  return a*k;
}

Point2 operator* (double k, const Point2 &a)
{
  return a*k;
}

Point2 randomPoint2 ()
{
  return Point2(randomNumber(-1.0, 1.0), randomNumber(-1.0, 1.0));
}

bool pointInCircumcircle (const Point2 &p, const Point2 &a, 
			  const Point2 &b, const Point2 &c)
{
  Parameter a1 = a.x - p.x, a2 = a.y - p.y, a3 = a1*a1 + a2*a2;
  Parameter b1 = b.x - p.x, b2 = b.y - p.y, b3 = b1*b1 + b2*b2;
  Parameter c1 = c.x - p.x, c2 = c.y - p.y, c3 = c1*c1 + c2*c2;
  Parameter k = a1*(b2*c3-b3*c2) - a2*(b1*c3-b3*c1) + a3*(b1*c2-b2*c1);
  return k.sign() == 1;
}

Point2 circumCenter (const Point2 &a, const Point2 &b, const Point2 &c)
{
  Parameter xy[2];
  linsolve2(a.y - b.y, c.y - a.y, b.x - a.x, a.x - c.x, 
	    0.5*(c.x - b.x), 0.5*(c.y - b.y), xy);
  return Point2(0.5*(a.x + b.x) + xy[0]*(a.y - b.y),
		0.5*(a.y + b.y) + xy[0]*(b.x - a.x));
}

void linsolve2 (Parameter a, Parameter b, Parameter c, Parameter d, 
		Parameter e, Parameter f, Parameter *xy)
{
  Parameter den = a*d - b*c;
  xy[0] = (e*d - b*f)/den;
  xy[1] = (a*f - c*e)/den;
}

Parameter signedArea (const Points2 &pts)
{
  Parameter a;
  int n = pts.size();
  for (int i = 0; i < n; ++i) {
    int j = (i + 1)%n;
    a = a + pts[i].crossproduct(pts[j]);
  }
  return a*0.5;
}

Point3 operator* (const Parameter &k, const Point3 &a)
{
  return a*k;
}

Point3 operator* (double k, const Point3 &a)
{
  return a*k;
}

Point3 segmentPlaneIntersectionPoint (const Point3 &e, const Point3 &u, 
				      const Point3 &n, const Point3 &a)
{
  Parameter nu = n.innerproduct(u), w = n.innerproduct(a-e), k = w/nu;
  return Point3(e.x + k*u.x, e.y + k*u.y, e.z + k*u.z);
}

Point3 lineIntersectionPoint (const Point3 &a, const Point3 &u,
			      const Point3 &b, const Point3 &v, int coord)
{
  Point2 aa = a.project(coord), uu = u.project(coord),
    bb = b.project(coord), vv = v.project(coord);
  Parameter k1 = uu.crossproduct(aa - bb), k2 = uu.crossproduct(vv), k = k1/k2;
  return Point3(b.x + k*v.x, b.y + k*v.y, b.z + k*v.z);
}

Point3 randomPoint3 ()
{
  return Point3(randomNumber(-1.0, 1.0), randomNumber(-1.0, 1.0),
		randomNumber(-1.0, 1.0));
}

ostream& operator<< (ostream &ostr, const Point2 &a)
{
  ostr << "(" << a.getX().round() << " " << a.getY().round() << ") ";
  return ostr;
}

void readPoints2 (Points2 &points, istream &istr)
{
  int n;
  double x, y;
  istr >> n;
  for (int i = 0; i < n; ++i) {
    istr >> x >> y;
    points.push_back(Point2(Parameter::input(x), Parameter::input(y)));
  }
}

void writePoints2 (const Points2 &points, ostream &ostr)
{
  ostr << setprecision(16) << points.size() << endl;
  for (unsigned int i = 0; i < points.size(); ++i)
    ostr << points[i].x.round() << " " << points[i].y.round() << endl;
}

ostream& operator<< (ostream &ostr, const Point3 &a)
{
  ostr << "(" << a.getX().round() << " " << a.getY().round()
       << " " << a.getZ().round() << ")";
  return ostr;
}

// debug

void pp (const Point2 &p)
{
  cerr << setprecision(16) << "(" << p.getX().round() << " " 
       << p.getY().round() << ")" << endl;
}

void pp (const Point3 &p)
{
  cerr << setprecision(16) << "(" << p.getX().round() << " " << p.getY().round()
       << " " << p.getZ().round() << ")" << endl;
}
