#include "ppoly.h"

PPoly::PPoly (int inv, int nti, Parameter *ia, int *im) : nv(inv), nt(nti)
{
  int k = nt*nv;
  a = new Parameter [nt];
  m = new int [k];
  for (int i = 0; i < nt; ++i)
    a[i] = ia[i];
  for (int i = 0; i < k; ++i)
    m[i] = im[i];
  setDegree();
}

PPoly::PPoly (const PPoly &p) : nv(p.nv), nt(p.nt)
{
  if (p.a) {
    int k = nt*nv;
    a = new Parameter [nt];
    m = new int [k];
    for (int i = 0; i < nt; ++i)
      a[i] = p.a[i];
    for (int i = 0; i < k; ++i)
      m[i] = p.m[i];
  }
  else {
    a = 0;
    m = 0;
  }
  d = p.d;
}

void PPoly::setDegree ()
{
  d = 0;
  for (int i = 0; i < nt; ++i) {
    int di = 0;
    for (int j = 0; j < nv; ++j)
      di += m[nv*i+j];
    if (d < di)
      d = di;
  }
}

PPoly & PPoly::operator= (const PPoly &p)
{
  nv = p.nv;
  nt = p.nt;
  delete [] a;
  delete [] m;
  if (p.a) {
    int k = nt*nv;
    a = new Parameter [nt];
    m = new int [k];
    for (int i = 0; i < nt; ++i)
      a[i] = p.a[i];
    for (int i = 0; i < k; ++i)
      m[i] = p.m[i];
  }
  else {
    a = 0;
    m = 0;
  }
  d = p.d;
  return *this;
}

Parameter PPoly::value (Parameter *x) const
{
  Parameter y;
  int *mp = m;
  for (int i = 0; i < nt; ++i) {
    Parameter z = a[i];
    for (int j = 0; j < nv; ++j, ++mp)
      for (int k = 0; k < *mp; ++k)
	z = z*x[j];
    y = y + z;
  }
  return y;
}

Parameter PPoly::der (Parameter *x, int dn) const
{
  Parameter y;
  int *mp = m;
  for (int i = 0; i < nt; ++i) 
    if (mp[dn] == 0)
      mp = mp + nv;
    else {
      Parameter z = a[i]*mp[dn];
      for (int j = 0; j < nv; ++j, ++mp)
	for (int k = 0; j == dn ? k < *mp -1 : k < *mp; ++k)
	  z = z*x[j];
      y = y + z;
    }
  return y;
}
 
PPoly PPoly::partial (int i) const
{
  Parameter fa[nt];
  int fm[nt*nv];
  int fnt = 0;
  for (int j = 0; j < nt; ++j)
    if (m[nv*j+i] > 0) {
      fa[fnt] = a[j]*m[j*nv+i];
      for (int k = 0; k < nv; ++k)
	fm[fnt*nv+k] = k == i ? m[j*nv+k] - 1 : m[j*nv+k];
      ++fnt;
    }
  return PPoly(nv, fnt, fa, fm);
}

PPoly PPoly::homogenize () const
{
  int mh[(nv+1)*nt];
  for (int i = 0; i < nt; ++i) {
    int dt = 0;
    for (int j = 0; j < nv; ++j) {
      dt += m[nv*i+j];
      mh[(nv+1)*i+j] = m[nv*i+j];
    }
    mh[(nv+1)*i+nv] = d - dt;
  }
  return PPoly(nv + 1, nt, a, mh);
}

void PPoly::increasePrecision ()
{
  for (int i = 0; i < nt; ++i)
    a[i].increasePrecision();
}

void PPoly::decreasePrecision ()
{
  for (int i = 0; i < nt; ++i)
    a[i].decreasePrecision();
}
