#include "solvecs.h"

int solveCS (const PPoly &f, Point2 *th)
{
  Parameter *t = new Parameter [4*f.degree()];
  int n1 = solveCSAux(f, true, t), n2 = solveCSAux(f, false, t + n1),
    n = n1 + n2;
  for (int i = 0; i < n; ++i)
    th[i] = pointCS(t[i], i < n1);
  delete [] t;
  extern long int nn1, nn2;
  if (f.a[0].extended())
    ++nn2;
  else
    ++nn1;
  return n;
}

int solveCSAux (const PPoly &f, bool flag, Parameter *t)
{
  int d = 2*f.degree();
  PFunPoly g = univariateCS(f, flag);
  Parameter *s = new Parameter [d];
  int n = implicit(g, -1.001, 1.001, s), m = 0;
  for (int i = 0; i < n; ++i)
    if (flag ? s[i].ub() > -1.0 && s[i].lb() < 1.0 :
	s[i].lb() >= -1.0 && s[i].ub() <= 1.0)
      t[m++] = s[i];
  delete [] s;
  return m;
}

PFunPoly univariateCS (const PPoly &f, bool flag)
{
  int d = f.degree(), dg = 0;
  Parameter *a = new Parameter[2*d+1];
  if (f.extended())
    for (int i = 0; i < 2*d + 1; ++i)
      a[i].increasePrecision();
  for (int i = 0; i < f.nt; ++i) {
    int c = f.m[2*i], s = f.m[2*i+1];
    termCS(f.a[i], c, s, d, flag, a);
    dg = max(dg, 2*d - s);
  }
  PFunPoly g(dg, a);
  delete [] a;
  return g;
}

void termCS (const Parameter &u, int c, int s, int d, bool flag, Parameter *a)
{
  int e = d - c - s;
  for (int k = 0; k <= c + e; ++k) {
    int l = 0;
    for (int i = 0; i <= k; ++i)
      l += power(-1, i)*choose(i, c)*choose(k - i, e);
    l *= flag || c%2 == 0 ? power(2, s) : - power(2, s);
    int p = 2*k + s;
    a[p] = a[p] + u*l;
  }
}

int choose (int k, int n)
{
  int res = 1;
  for (int i = 0; i < k; ++i)
    res *= n - i;
  for (int i = 2; i <= k; ++i)
    res /= i;
  return res;
}

int power (int b, int d)
{
  int p = 1;
  for (int i = 0; i < d; ++i)
    p *= b;
  return p;
}

Point2 pointCS (const Parameter &t, bool flag)
{
  Parameter x = flag ? 1.0 - t*t : t*t - 1.0, y = 2.0*t,
    d = 1.0 + t*t;
  return Point2(x/d, y/d);
}

Point2 polishCS (const PPoly &f, Point2 &th)
{
  Parameter t = th.y/(1.0 + th.x);
  bool flag = fabs(t.round()) < 100.0;
  if (!flag)
    t = th.y/(1.0 - th.x);
  PFunPoly g = univariateCS(f, flag);
  Parameter pt = g.polish(t);
  return pointCS(pt, flag);  
}
