#include "sum.h"

int IdCounter = 0;

bool Vertex::orderZ (Vertex *v)
{
  if (this == v)
    return false;
  bool flag;
  try {
    flag = p.z < v->p.z;
  }
  catch (SignException &exc) {
    increasePrecision();
    v->increasePrecision();
    flag = p.z < v->p.z;
    decreasePrecision();
    v->decreasePrecision();
  }
  return flag;
}

bool Vertex::steeperZ (Vertex *v, Vertex *w)
{
  bool flag;
  try {
    flag = v->cosZ(this) < w->cosZ(this);
  }
  catch (SignException &exc) {
    increasePrecision();
    v->increasePrecision();
    w->increasePrecision();
    flag = v->cosZ(this) < w->cosZ(this);
    decreasePrecision();
    v->decreasePrecision();
    w->decreasePrecision();
  }
  return flag;
}

Parameter Vertex::cosZ (Vertex *v) const
{
  return (v->p.z - p.z)/v->p.distance(p);
}

bool PVertex::compatible (PFace *f) const
{
  for (int i = 0; i < edges.size(); ++i)
    if (edges[i]->u().innerproduct(f->n).sign() == 1)
      return false;
  return true;
}

bool PVertex::onFace (PFace *f) const
{ 
  return fids.find(f->id) != fids.end();
}

void PVertex::increasePrecision () 
{ 
  if (parent) {
    parent->increasePrecision();
    p = parent->p;
  }
  else
    p.increasePrecision(); 
}

void PVertex::decreasePrecision () 
{
  if (parent)
    parent->decreasePrecision();
  p.decreasePrecision(); 
}

void PVertex::increasePrecisionCone () 
{
  increasePrecision();
  for (PEdges::iterator e = edges.begin(); e != edges.end(); ++e)
    (*e)->face->increasePrecision();
}

void PVertex::decreasePrecisionCone () 
{
  decreasePrecision();
  for (PEdges::iterator e = edges.begin(); e != edges.end(); ++e)
    (*e)->face->decreasePrecision();
}

void PVertexD::increasePrecision () 
{
  a->increasePrecision();
  b->increasePrecision();
  p = a->p - b->p;
}

void PVertexD::decreasePrecision () 
{
  a->decreasePrecision();
  b->decreasePrecision();
  p.decreasePrecision();
}

bool HFace::conflict (PEdge *f) const
{
  return f->coplanar(e, next->e) ||
    f->u().circulation(e->u(), next->e->u()).sign() == 1;
}

void HFace::updateCset (HFace *h, PEdge *f)
{
  for (PEdges::iterator g = h->cset.begin(); g != h->cset.end(); ++g)
    if (*g != f && conflict(*g))
      cset.push_back(*g);
}

void HFace::cone (Points3 &g) const
{
  const HFace *h = this;
  do {
    g.push_back(h->e->u().crossproduct(h->next->e->u()));
    h = h->next;
  }
  while (h != this);
}

void PVertexCone::increasePrecision ()
{
  for (Points3::iterator p = g.begin(); p != g.end(); ++p)
    p->increasePrecision();
}

void PVertexCone::decreasePrecision ()
{
  for (Points3::iterator p = g.begin(); p != g.end(); ++p)
    p->decreasePrecision();
}

PVertexCone * convexCone (PVertex *v)
{
  PEdges edges;
  for (PEdges::const_iterator e = v->edges.begin(); e != v->edges.end(); ++e)
    if ((*e)->convex())
      edges.push_back(*e);
  if (edges.size() < 3)
    return 0;
  HFace *hull = initHull(edges);
  if (!hull)
    return 0;
  for (int i = 3; i < edges.size(); ++i) {
    bool flag;
    hull = updateHull(hull, edges[i], flag);
    if (!flag) {
      deleteHull(hull);
      return 0;
    }
  }
  PVertexCone *c = new PVertexCone(v);
  hull->cone(c->g);
  deleteHull(hull);
  return c;
}

HFace * initHull (PEdges &edges)
{
  int k = 2;
  while (k < edges.size() && edges[0]->coplanar(edges[1], edges[k]))
    ++k;
  if (k == edges.size())
    return 0;
  if (k > 2) {
    PEdge *temp = edges[2];
    edges[2] = edges[k];
    edges[k] = temp;
  }
  PEdge *e = edges[0], *f, *g;
  if (e->u().circulation(edges[1]->u(), edges[2]->u()).sign() == -1) {
    f = edges[1];
    g = edges[2];
  }
  else {
    f = edges[2];
    g = edges[1];
  }
  HFace *he = new HFace(e, 0, 0), *hf = new HFace(f, he, 0), 
    *hg = new HFace(g, hf, he);
  he->next = hf;
  he->prev = hg;
  hf->next = hg;
  HFace *faces[3] = {he, hf, hg};
  for (int i = 0; i < 3; ++i)
    for (int j = 3; j < edges.size(); ++j)
      if (faces[i]->conflict(edges[j]))
	faces[i]->cset.push_back(edges[j]);
  return he;
}

HFace * updateHull (HFace *hull, PEdge *e, bool &flag)
{
  flag = true;
  HFace *fs = hull;
  if (fs->inCset(e)) {
    while (fs->prev != hull && fs->prev->inCset(e))
      fs = fs->prev;
    if (fs->prev == hull) {
      flag = false;
      return hull;
    }
  }
  else {
    do
      fs = fs->next;
    while (fs != hull && !fs->inCset(e));
      if (fs == hull)
	return hull;
  }
  HFace *fe = fs;
  while (fe->next->inCset(e))
    fe = fe->next;
  return updateHullAux(fs, fe, e);
}

HFace * updateHullAux (HFace *fs, HFace *fe, PEdge *e)
{
  HFace *f1 = new HFace(fs->e, fs->prev, 0), *f2 = new HFace(e, f1, fe->next);
  f1->next = f2;
  f1->updateCset(fs->prev, e);
  f1->updateCset(fs, e);
  f2->updateCset(fe, e);
  f2->updateCset(fe->next, e);
  f1->prev->next = f1;
  f2->next->prev = f2;
  fe->next = 0;
  while (fs) {
    HFace *ptr = fs->next;
    delete fs;
    fs = ptr;
  }
  return f1;
}

void deleteHull (HFace *hull)
{
  hull->prev->next = 0;
  while (hull) {
    HFace *nhull = hull->next;
    delete hull;
    hull = nhull;
  }
}

bool PEdge::convex ()
{
  PEdge *e = face->boundary[(faceIndex() + 2)%3]->twin,
    *f = twin->face->boundary[(twin->faceIndex() + 1)%3];
  return !coplanar(e, f) &&
    face->n.circulation(u(), twin->face->n).sign() == -1;
}

bool PEdge::compatible (PEdge *f, bool &aflag)
{
  try {
    return compatibleAux(f, aflag);
  }
  catch (SignException &exc) {
    increasePrecision();
    f->increasePrecision();
    face->increasePrecision();
    twin->face->increasePrecision();
    f->face->increasePrecision();
    f->twin->face->increasePrecision();
    bool flag = compatibleAux(f, aflag);
    decreasePrecision();
    f->decreasePrecision();
    face->decreasePrecision();
    twin->face->decreasePrecision();
    f->face->decreasePrecision();
    f->twin->face->decreasePrecision();
    return flag;
  }
}

bool PEdge::compatibleAux (PEdge *f, bool &aflag) const
{
  Point3 fu = f->u();
  int s1 = face->n.innerproduct(fu).sign(),
    s2 = twin->face->n.innerproduct(fu).sign();
  aflag = s1 == 1;
  if (s1 == s2)
    return false;
  Point3 eu = u();
  int s3 = f->face->n.innerproduct(eu).sign();
  if (s1 == s3)
    return false;
  int s4 = f->twin->face->n.innerproduct(eu).sign();
  return s1 == s4;
}

void PEdge::fids (set<int> &res) const
{
  set_intersection(tail->fids.begin(), tail->fids.end(),
		   head()->fids.begin(), head()->fids.end(),
		   inserter(res, res.begin()));
}

bool PEdge::coplanar (PEdge *e) const
{
  return face->id == e->face->id || face->id == e->twin->face->id ||
    twin->face->id == e->face->id || twin->face->id == e->twin->face->id;
}

bool PEdge::coplanar (PEdge *e, PEdge *f) const
{
  if (colinear(e->head()) || colinear(f->head()) || e->colinear(f->head()))
    return true;
  set<int> &sdh = head()->fids, &seh = e->head()->fids, &sfh = f->head()->fids, 
    sefh, sdefh;
  set_intersection(seh.begin(), seh.end(), sfh.begin(), sfh.end(),
		   inserter(sefh, sefh.begin()));
  set_intersection(sdh.begin(), sdh.end(), sefh.begin(), sefh.end(),
		   inserter(sdefh, sdefh.begin()));
  if (sdefh.size() == 2)
    return true;
  set<int> &sdt = tail->fids, sall;
  set_intersection(sdt.begin(), sdt.end(), sdefh.begin(), sdefh.end(),
		   inserter(sall, sall.begin()));
  return !sall.empty();
}

bool PEdge::colinear (const PVertex *v) const
{
  return v == tail || v == head() ||
    face->id != twin->face->id && v->onFace(face) && v->onFace(twin->face);
}

bool PEdge::colinear (PEdge *e) const
{
  return colinear(e->tail) && colinear(e->head());
}

bool PEdge::acute () const
{
  PVertex *a = oppositeVertex();
  Point3 v = a->p - tail->p, eu = u();
  Parameter uu = eu.normSquared(), uv = eu.innerproduct(v);
  if (uv.sign() == -1)
    return false;
  return uv < uu;
}

PVertex * PEdge::oppositeVertex () const
{
  return face->boundary[(faceIndex()+2)%3]->tail;
}

int PEdge::faceIndex () const
{
  for (int i = 0; i < 3; ++i)
    if (face->boundary[i] == this)
      return i;
  return -1;
}

void PEdge::increasePrecision ()
{ 
  tail->increasePrecision(); 
  head()->increasePrecision(); 
}

void PEdge::decreasePrecision ()
{ 
  tail->decreasePrecision(); 
  head()->decreasePrecision(); 
}

bool PFace::sharedEdge (PFace *g) const
{
  for (int i = 0; i < 3; ++i)
    if (boundary[i]->twin->face->id == g->id ||
	g->boundary[i]->twin->face->id == id)
      return true;
  return false;
}

bool PFace::rayIntersection (const Point3 &a, const Point3 &u, Parameter &k) const
{
  k = n.innerproduct(boundary[0]->tail->p - a)/n.innerproduct(u);
  if (k.sign() == -1)
    return false;
  int coord = n.projectionCoordinate();
  Point3 p = a + k*u;
  Point2 q = p.project(coord);
  for (int i = 0; i < 3; ++i) {
    Point2 pc = boundary[i]->tail->p.project(coord), 
      pv = boundary[i]->u().project(coord);
    if (pv.crossproduct(q - pc).sign() == -1)
      return false;
  }
  return true;
}

void PFace::increasePrecision ()
{ 
  for (int i = 0; i < 3; ++i)
    boundary[i]->increasePrecision();
  n = boundary[0]->u().crossproduct(boundary[1]->u());
}

void PFace::decreasePrecision ()
{ 
  for (int i = 0; i < 3; ++i)
    boundary[i]->decreasePrecision();
  n.decreasePrecision();
}

bool PEdit::operator< (PEdit &b)
{
  if (valid && !b.valid)
    return true;
  if (!valid)
    return false;
  bool res;
  try {
    res = d < b.d;
  }
  catch (SignException &exc) {
    increasePrecision();
    b.increasePrecision();
    try {
      res = d < b.d;
    }
    catch (SignException &exc) {
      res = d.lb() < b.d.lb();
    }
    decreasePrecision();
    b.decreasePrecision();
  }
  return res;
}

void PEdit::increasePrecision ()
{
  edge->increasePrecision();
  if (cflag)
    d = edge->u().normSquared();
  else {
    PVertex *c = edge->oppositeVertex();
    c->increasePrecision();
    Point3 &a = edge->tail->p, &b = edge->head()->p, u = b - a, v = c->p - a;
    Parameter uu = u.normSquared(), uv = u.innerproduct(v);
    d = v.normSquared() - uv*uv/uu;
    c->decreasePrecision();
  }
  edge->decreasePrecision();
}

void PEdit::decreasePrecision ()
{
  d.decreasePrecision();
}

bool PEditOrder::operator() (PEdit *a, PEdit *b) const
{
  if (a == b)
    return false;
  bool flag;
  try {
    flag = *b < *a;
  }
  catch (SignException &exc) {
    a->increasePrecision();
    b->increasePrecision();
    flag = *b < *a;
    a->decreasePrecision();
    b->decreasePrecision();
  }
  return flag;
}

PEdit * nextPEdit (PEdits &edits)
{
  while (!edits.empty()) {
    PEdit *e = edits[0];
    pop_heap(edits.begin(), edits.end(), PEditOrder());
    edits.pop_back();
    if (e->valid)
      return e;
    delete e;
  }
  return 0;
}

Parameter minAltitude (PVertex *a, PVertex *b, PVertex *c, 
		       PVertex *d, bool &cflag)
{
  Point3 u = b->p - a->p, v = c->p - a->p, w = d->p - a->p;
  Parameter uu = u.normSquared(), uv = u.innerproduct(v),
    uw = u.innerproduct(w), hc = v.normSquared() - uv*uv/uu,
    hd = w.normSquared() - uw*uw/uu;
  try {
    cflag = hc < hd;
  }
  catch (SignException &exc) {
    if (hc.extended())
      cflag = hc.lb() < hd.lb();
    else
      throw signException;
  }
  return cflag ? hc : hd;
}

Polyhedron::~Polyhedron ()
{
  for (PVertices::iterator v = vertices.begin(); v != vertices.end(); ++v)
    delete *v;
  for (PEdges::iterator e = edges.begin(); e != edges.end(); ++e) {
    delete (*e)->twin;
    delete *e;
  }
  for (PFaces::iterator f = faces.begin(); f != faces.end(); ++f)
    delete *f;
}

PVertex * Polyhedron::addVertex (const Point3 &p)
{
  PVertex *v = new PVertex(p);
  addVertex(v);
  return v;
}

void Polyhedron::addVertex (PVertex *v)
{
  const Point3 &p = v->p;
  if (vertices.empty()) {
    bbox[0] = p.getX().lb();
    bbox[1] = p.getX().ub();
    bbox[2] = p.getY().lb();
    bbox[3] = p.getY().ub();
    bbox[4] = p.getZ().lb();
    bbox[5] = p.getZ().ub();
  }
  else {
    bbox[0] = min(bbox[0], p.getX().lb());
    bbox[1] = max(bbox[1], p.getX().ub());
    bbox[2] = min(bbox[2], p.getY().lb());
    bbox[3] = max(bbox[3], p.getY().ub());
    bbox[4] = min(bbox[4], p.getZ().lb());
    bbox[5] = max(bbox[5], p.getZ().ub());
  }
  vertices.push_back(v);
}

PEdge * Polyhedron::getEdge (PVertex *a, PVertex *b)
{
  for (PEdges::iterator e = a->edges.begin(); e != a->edges.end(); ++e)
    if ((*e)->head() == b)
      return *e;
  PEdge *eab = new PEdge(a), *eba = new PEdge(b, eab);
  eab->twin = eba;
  edges.push_back(eab);
  a->edges.push_back(eab);
  b->edges.push_back(eba);
  return eab;
}

void Polyhedron::addFace (PVertex *a, PVertex *b, PVertex *c, int id)
{
  PFace *f = new PFace;
  f->id = id;
  f->boundary[0] = getEdge(a, b);
  f->boundary[1] = getEdge(b, c);
  f->boundary[2] = getEdge(c, a);
  f->boundary[0]->face = f;
  f->boundary[1]->face = f;
  f->boundary[2]->face = f;
  f->n = f->boundary[0]->u().crossproduct(f->boundary[1]->u());
  faces.push_back(f);
}

void Polyhedron::convexEdges (PEdges &convex) const
{
  for (PEdges::const_iterator e = edges.begin(); e != edges.end(); ++e) {
    bool flag;
    try {
      flag = (*e)->convex();
    }
    catch (SignException &excp) {
      (*e)->increasePrecision();
      (*e)->face->increasePrecision();
      (*e)->twin->face->increasePrecision();
      flag = (*e)->convex();
      (*e)->decreasePrecision();
      (*e)->face->decreasePrecision();
      (*e)->twin->face->decreasePrecision();
    }
    if (flag)
      convex.push_back(*e);
  }
}

Polyhedron * Polyhedron::translate (const Point3 &t) const
{
  Polyhedron *p = new Polyhedron;
  PVPVMap vmap;
  for (PVertices::const_iterator v = vertices.begin(); v != vertices.end(); ++v) {
    PVertex *w = p->addVertex((*v)->p + t);
    vmap.insert(pair<PVertex *, PVertex *>(*v, w));
  }
  for (PFaces::const_iterator f = faces.begin(); f != faces.end(); ++f) {
    PVPVMap::iterator a = vmap.find((*f)->boundary[0]->tail),
      b = vmap.find((*f)->boundary[1]->tail), c = vmap.find((*f)->boundary[2]->tail);
    p->addFace(a->second, b->second, c->second);
  }
  return p;
}

Polyhedron * Polyhedron::rotate (const Quat &q) const
{
  Polyhedron *p = new Polyhedron;
  PVPVMap vmap;
  for (PVertices::const_iterator v = vertices.begin(); v != vertices.end(); ++v) {
    PVertex *w = p->addVertex(q*(*v)->p);
    vmap.insert(pair<PVertex *, PVertex *>(*v, w));
  }
  for (PFaces::const_iterator f = faces.begin(); f != faces.end(); ++f) {
    PVPVMap::iterator a = vmap.find((*f)->boundary[0]->tail),
      b = vmap.find((*f)->boundary[1]->tail), c = vmap.find((*f)->boundary[2]->tail);
    p->addFace(a->second, b->second, c->second);
  }
  return p;
}

Polyhedron * Polyhedron::negative () const
{
  Polyhedron *p = new Polyhedron;
  PVPVMap vmap;
  for (PVertices::const_iterator v = vertices.begin(); v != vertices.end(); ++v) {
    PVertex *w = p->addVertex(- (*v)->p);
    vmap.insert(pair<PVertex *, PVertex *>(*v, w));
  }
  for (PFaces::const_iterator f = faces.begin(); f != faces.end(); ++f) {
    PVPVMap::iterator a = vmap.find((*f)->boundary[0]->tail),
      b = vmap.find((*f)->boundary[1]->tail), c = vmap.find((*f)->boundary[2]->tail);
    p->addFace(c->second, b->second, a->second);
  }
  return p;
}

Polyhedron * Polyhedron::negativeTranslate (const Point3 &t) const
{
  Polyhedron *p = new Polyhedron;
  PVPVMap vmap;
  for (PVertices::const_iterator v = vertices.begin(); v != vertices.end(); ++v) {
    PVertex *w = p->addVertex(t - (*v)->p);
    vmap.insert(pair<PVertex *, PVertex *>(*v, w));
  }
  for (PFaces::const_iterator f = faces.begin(); f != faces.end(); ++f) {
    PVPVMap::iterator a = vmap.find((*f)->boundary[0]->tail),
      b = vmap.find((*f)->boundary[1]->tail), c = vmap.find((*f)->boundary[2]->tail);
    p->addFace(c->second, b->second, a->second);
  }
  return p;
}

bool Polyhedron::intersects (const Polyhedron *b, PVertex *at, PVertex *bt) const
{
  if (!bboxOverlap(bbox, b->bbox))
    return false;
  PVertex *a0 = vertices[0] == at ? vertices[1] : vertices[0],
    *b0 = b->vertices[0] == bt ? b->vertices[1] : b->vertices[0];
  if (contains(b0) || b->contains(a0))
    return true;
  Convolution con;
  con.merge(this, true, false);
  con.merge(b, false, false);
  return con.testIntersection(at, bt);
}

bool Polyhedron::contains (PVertex *v) const
{
  if (!bboxContains(bbox, v->p))
    return false;
  bool res = false;
  Parameter k;
  Point3 u = randomPoint3();
  for (PFaces::const_iterator f = faces.begin(); f != faces.end(); ++f) {
    bool flag;
    try {
      flag = (*f)->rayIntersection(v->p, u, k);
    }
    catch (SignException &exc) {
      (*f)->increasePrecision();
      v->increasePrecision();
      u.increasePrecision();
      flag = (*f)->rayIntersection(v->p, u, k);
      (*f)->decreasePrecision();
      v->decreasePrecision();
      u.decreasePrecision();
    }
    if (flag)
      res = !res;
  }
  return res;
}

Point3 Polyhedron::interiorPoint () const
{
  PFace *g = largestFace();
  PEdge **b = g->boundary;
  Point3 a = (b[0]->tail->p + b[1]->tail->p + b[2]->tail->p)*(1.0/3.0),
    n = - g->n;
  bool flag = true;
  Parameter k, kf;
  for (PFaces::const_iterator f = faces.begin(); f != faces.end(); ++f)
    if (!(g == *f || g->id != -1 && g->id == (*f)->id)) {
      bool iflag;
      try {
	iflag = (*f)->rayIntersection(a, n, kf);
      }
      catch (SignException &exc) {
	(*f)->increasePrecision();
	g->increasePrecision();
	a = (b[0]->tail->p + b[1]->tail->p + b[2]->tail->p)*(1.0/3.0);
	n = - g->n;
	iflag = (*f)->rayIntersection(a, n, kf);
	kf.decreasePrecision();
	(*f)->decreasePrecision();
	g->decreasePrecision();
	a = (b[0]->tail->p + b[1]->tail->p + b[2]->tail->p)*(1.0/3.0);
	n = - g->n;
      }
      if (iflag && (flag || kf < k)) {
	flag = false;
	k = kf;
      }
    }
  return (a + 0.5*k*n).round();
}

PFace * Polyhedron::largestFace () const
{
  double amax = -1.0;
  PFace *fmax = 0;
  for (PFaces::const_iterator f = faces.begin(); f != faces.end(); ++f) {
    int coord = (*f)->n.projectionCoordinate();
    Point3 u = (*f)->boundary[0]->u(), v = (*f)->boundary[1]->u();
    double af = u.crossproduct(v, coord).round();
    if (amax < af) {
      amax = af;
      fmax = *f;
    }
  }
  return fmax;
}

void Polyhedron::simplify (double dmin)
{
  double dmin2 = dmin*dmin;
  PEdits edits;
  simplifyInit(dmin2, edits);
  while (true) {
    PEdit *e = nextPEdit(edits);
    if (!e)
      break;
    if (e->cflag)
      collapseEdge(e->edge, dmin2, edits);
    else
      flipEdge(e->edge, dmin2, edits);
    delete e;
  }
  simplifyEnd();
}

void Polyhedron::simplifyInit (double dmin, PEdits &edits) const
{
  for (PEdges::const_iterator e = edges.begin(); e != edges.end(); ++e)
    checkEdits(*e, dmin, edits);
}

void Polyhedron::checkEdits (PEdge *e, double dmin, PEdits &edits) const
{
  checkCollapse(e, dmin, edits);
  checkFlip(e, dmin, edits);
}

void Polyhedron::checkCollapse (PEdge *e, double dmin, PEdits &edits) const
{
  Parameter d = e->u().normSquared();
  bool flag = false;
  try {
    flag = dmin < d;
  }
  catch (SignException &exc) {
    e->increasePrecision();
    flag = dmin < e->u().normSquared();
    e->decreasePrecision();
  }
  if (flag) return;
  PEdit *edit = new PEdit(true, e, d);
  e->cedit = edit;
  edits.push_back(edit);
  push_heap(edits.begin(), edits.end(), PEditOrder());
}

void Polyhedron::checkFlip (PEdge *e, double dmin, PEdits &edits) const
{
  bool flip, cflag;
  Parameter h;
  try {
    flip = checkFlipAux(e, dmin, h, cflag);
  }
  catch (SignException &exc) {
    e->increasePrecision();
     e->oppositeVertex()->increasePrecision();
    e->twin->oppositeVertex()->increasePrecision();
    try {
      flip = checkFlipAux(e, dmin, h, cflag);
    }
    catch (SignException &exc) {
      flip = false;
    }
    e->decreasePrecision();
    e->oppositeVertex()->decreasePrecision();
    e->twin->oppositeVertex()->decreasePrecision();
    h.decreasePrecision();
  }
  if (!flip)
    return;
  PEdit *edit = new PEdit(false, cflag ? e : e->twin, h);
  if (cflag)
    e->fedit = edit;
  else
    e->twin->fedit = edit;
  edits.push_back(edit);
  push_heap(edits.begin(), edits.end(), PEditOrder());
}

bool Polyhedron::checkFlipAux (PEdge *e, double dmin, Parameter &h, 
			       bool &cflag) const
{
  if (!(e->acute() && e->twin->acute()))
    return false;
  PVertex *a = e->tail, *b = e->head(), *c = e->oppositeVertex(),
    *d = e->twin->oppositeVertex();
  h = minAltitude(a, b, c, d, cflag);
  if (dmin < h)
    return false;
  bool ncflag;
  Parameter nh = minAltitude(a, d, c, b, ncflag);
  return h < nh;
}

void Polyhedron::collapseEdge (PEdge *e, double dmin, PEdits &edits)
{
  int ei = e->faceIndex(), eti = e->twin->faceIndex();
  PEdge *e1 = e->face->boundary[(ei+1)%3], 
    *e2 = e->face->boundary[(ei+2)%3],
    *e3 = e->twin->face->boundary[(eti+1)%3], 
    *e4 = e->twin->face->boundary[(eti+2)%3];
  PVertex *a = e->tail, *b = e->head(), *c = e2->tail, *d = e4->tail;
  if (e1->twin->face == e4->twin->face &&
      e2->twin->face == e3->twin->face) {
    invalidateCollapse(e);
    return;
    }
  for (PEdges::iterator x = a->edges.begin(); x != a->edges.end(); ++x) {
    PVertex *xh = (*x)->head();
    if (xh != c && xh != d)
      for (PEdges::iterator y = xh->edges.begin(); y != xh->edges.end(); ++y)
	if ((*y)->head() == b) {
	  invalidateCollapse(e);
	  return;
	}
  }
  e->face->boundary[0] = 0;
  e->twin->face->boundary[0] = 0;
  removeEdge(e); 
  if (e1->twin->face == e4->twin->face) {
    e1->twin->face->boundary[0] = 0;
    removeEdge(e1);
    removeEdge(e4);
    addFace(a, d, c);
    mergeVertices(b, a);
    updateEdits(a, true, dmin, edits);
  }
  else if (e2->twin->face == e3->twin->face) {
    e2->twin->face->boundary[0] = 0;
    removeEdge(e2);
    removeEdge(e3);
    addFace(b, c, d);
    mergeVertices(a, b);
    updateEdits(b, true, dmin, edits);
  }
  else {
    e1->face = e2->twin->face;
    e1->face->boundary[e2->twin->faceIndex()] = e1;
    e4->face = e3->twin->face;
    e4->face->boundary[e3->twin->faceIndex()] = e4;
    removeEdge(e2);
    removeEdge(e3);
    mergeVertices(a, b);
    updateEdits(b, true, dmin, edits);
  }
}

void Polyhedron::mergeVertices (PVertex *a, PVertex *b) const
{
  b->p = (0.5*(a->p + b->p)).round();
  for (PEdges::iterator x = a->edges.begin(); x != a->edges.end(); ++x) {
    (*x)->tail = b;
    b->edges.push_back(*x);
  }
  a->edges.clear();
  for (PEdges::iterator e = b->edges.begin(); e != b->edges.end(); ++e) {
    Point3 eu = (*e)->face->boundary[0]->u(), ev = (*e)->face->boundary[1]->u();
    (*e)->face->n = eu.crossproduct(ev);
  }
}

void Polyhedron::removeEdge (PEdge *e) const
{
  PVertex *t = e->tail, *h = e->head();
  PEdges::iterator iter = remove(t->edges.begin(), t->edges.end(), e);
  t->edges.erase(iter);
  iter = remove(h->edges.begin(), h->edges.end(), e->twin);
  h->edges.erase(iter);
  e->face = e->twin->face = 0;
  invalidateEdits(e);
}

void Polyhedron::flipEdge (PEdge *e, double dmin, PEdits &edits)
{
  int ei = e->faceIndex(), eti = e->twin->faceIndex();
  PEdge *e1 = e->face->boundary[(ei+1)%3], 
    *e2 = e->face->boundary[(ei+2)%3],
    *e3 = e->twin->face->boundary[(eti+1)%3], 
    *e4 = e->twin->face->boundary[(eti+2)%3];
  PVertex *a = e->tail, *b = e->head(), *c = e2->tail, *d = e4->tail;
  for (PEdges::iterator x = c->edges.begin(); x != c->edges.end(); ++x)
    if ((*x)->head() == d) {
      invalidateFlip(e);
      return;
    }
  e->face->boundary[0] = e->twin->face->boundary[0] = 0;
  removeEdge(e);
  addFace(a, d, c);
  addFace(b, c, d);
  checkCollapse(getEdge(c, d), dmin, edits);
  updateEdits(a, false, dmin, edits);
  updateEdits(b, false, dmin, edits);
}

void Polyhedron::invalidateEdits (PEdge *e) const
{
  invalidateCollapse(e);
  invalidateFlip(e);
}

void Polyhedron::invalidateCollapse (PEdge *e) const
{
  if (e->cedit)
    e->cedit->valid = false;
  if (e->twin->cedit)
    e->twin->cedit->valid = false;
  e->cedit = e->twin->cedit = 0;
}

void Polyhedron::invalidateFlip (PEdge *e) const
{
  if (e->fedit)
    e->fedit->valid = false;
  if (e->twin->fedit)
    e->twin->fedit->valid = false;
  e->fedit = e->twin->fedit = 0;
}

void Polyhedron::updateEdits (PVertex *v, bool oflag, double dmin, 
			      PEdits &edits) const
{
  for (PEdges::iterator e = v->edges.begin(); e != v->edges.end(); ++e) {
    invalidateEdits(*e);
    checkEdits(*e, dmin, edits);
    if (oflag) {
      PEdge *f = (*e)->face->boundary[((*e)->faceIndex()+1)%3];
      invalidateEdits(f);
      checkEdits(f, dmin, edits);
    }
  }
}

void Polyhedron::simplifyEnd ()
{
  int i = 0;
  while (i < vertices.size())
    if (vertices[i]->edges.empty()) {
      delete vertices[i];
      vertices[i] = *(vertices.end()-1);
      vertices.pop_back();
    }
    else {
      vertices[i]->p = vertices[i]->p.round();
      vertices[i]->parent = 0;
      ++i;
    }
  i = 0;
  while (i < edges.size())
    if (!edges[i]->face) {
      delete edges[i]->twin;
      delete edges[i];
      edges[i] = *(edges.end()-1);
      edges.pop_back();
    }
    else
      ++i;
  i = 0;
  while (i < faces.size())
    if (!faces[i]->boundary[0]) {
      delete faces[i];
      faces[i] =  *(faces.end()-1);
      faces.pop_back();
    }
    else {
      faces[i]->n = faces[i]->n.round();
      ++i;
    }
}

double Polyhedron::maxBBoxDimension () const
{
  return max(bbox[1] - bbox[0], 
	     max(bbox[3] - bbox[2], bbox[5] - bbox[4]));
}

bool Polyhedron::freeSpace (double *box, double *res) const
{
  for (int i = 0; i < 6; ++i)
    res[i] = box[i] - bbox[i];
  return res[0] < res[1] && res[2] < res[3] && res[4] < res[5];
}

void Polyhedron::output (PFormat format, ostream &ostr) const
{
  VIMap vimap;
  Points3 pts;
  vector<int> data;
  for (PFaces::const_iterator f = faces.begin(); f != faces.end(); ++f)
    for (int i = 0; i < 3; ++i)
      data.push_back(getPoint(vimap, pts, (*f)->boundary[i]->tail));
  switch (format) {
  case VTK:
    outputVTK(pts, data, ostr);
    break;
  case OFF:
    outputOFF(pts, data, ostr);
    break;
  case TRI:
    outputTRI(pts, data, ostr);
    break;
  }
}

bool bboxContains (const double *bbox, const Point3 &p)
{
  return p.x.lb() <= bbox[1] && p.x.ub() >= bbox[0] &&
    p.y.lb() <= bbox[3] && p.y.ub() >= bbox[2] &&
    p.z.lb() <= bbox[5] && p.z.ub() >= bbox[4];
}

int getPoint (VIMap &vimap, Points3 &pts, Vertex *v)
{
  VIMap::iterator iter = vimap.find(v);
  if (iter != vimap.end())
    return iter->second;
  int k = pts.size();
  pts.push_back(v->p);
  vimap.insert(pair<Vertex *, int>(v, k));
  return k;
}

void outputVTK (const Points3 &pts, const vector<int> &data, ostream &ostr)
{
  int np = pts.size(), nt = data.size()/3;
  ostr << setprecision(16) << "# vtk DataFile Version 3.0" << endl
       << "vtk output" << endl << "ASCII" << endl
       << "DATASET POLYDATA" << endl 
       << "POINTS " << np << " double" << endl;
  for (Points3::const_iterator p = pts.begin(); p != pts.end(); ++p) 
    ostr << p->getX().round() << " " << p->getY().round() << " " 
	 << p->getZ().round() << endl;
  ostr << endl << "POLYGONS " << nt << " " << 4*nt << endl;
  for (int i = 0; i < nt; ++i)
    ostr << "3 " << data[3*i] << " " << data[3*i+1] << " "
	 << data[3*i+2] << endl;
  ostr << endl << "CELL_DATA " << nt << endl 
       << "POINT_DATA " << np << endl;
}

void outputOFF (const Points3 &pts, const vector<int> &data, ostream &ostr)
{
  int np = pts.size(), nt = data.size()/3;
  ostr << setprecision(16) << "OFF" << endl << np << " " << nt << " 0" << endl;
  for (Points3::const_iterator p = pts.begin(); p != pts.end(); ++p) 
    ostr << p->getX().round() << " " << p->getY().round() << " " 
	 << p->getZ().round() << endl;
  for (int i = 0; i < nt; ++i)
    ostr << "3 " << data[3*i] << " " << data[3*i+1] << " "
	 << data[3*i+2] << endl;
}

void outputTRI (const Points3 &pts, const vector<int> &data, ostream &ostr)
{
  int np = pts.size(), nt = data.size()/3;
  ostr << setprecision(16) << np << " " << nt << endl;
  for (Points3::const_iterator p = pts.begin(); p != pts.end(); ++p) 
    ostr << p->getX().round() << " " << p->getY().round() << " " 
	 << p->getZ().round() << endl;
  for (int i = 0; i < nt; ++i)
    ostr << data[3*i] << " " << data[3*i+1] << " "
	 << data[3*i+2] << endl;
}

Polyhedron * readPolyhedron (istream &istr)
{
  int nv, nt, *t;
  double *v;
  char ch = istr.peek();
  if (ch == '#')
    readPolyhedronVTK(istr, nv, v, nt, t);
  else if (ch == 'O' || ch == 'o')
    readPolyhedronOFF(istr, nv, v, nt, t);
  else
    readPolyhedronTRI(istr, nv, v, nt, t);
  Polyhedron *p = polyhedron(nv, v, nt, t);
  delete [] v;
  delete [] t;
  return p;
}

void readPolyhedronVTK (istream &istr, int &nv, double *&v, int &nt, int *&t)
{
  string dummy;
  int idummy, k;
  for (int i = 0; i < 11; ++i)
    istr >> dummy;
  istr >> nv >> dummy;
  v = new double [3*nv];
  for (int i = 0; i < nv; ++i)
    istr >> v[3*i] >> v[3*i+1] >> v[3*i+2];
  istr >> dummy >> nt >> idummy;
  t = new int[3*nt];
  for (int i = 0; i < nt; ++i) {
    istr >> k >> t[3*i] >> t[3*i+1] >> t[3*i+2];
    assert(k == 3);
  }
  istr >> dummy >> idummy;
}

void readPolyhedronOFF (istream &istr, int &nv, double *&v, int &nt, int *&t)
{
  string dummy;
  int ne, k;
  istr >> dummy >> nv >> nt >> ne;
  v = new double [3*nv];
  for (int i = 0; i < nv; ++i)
    istr >> v[3*i] >> v[3*i+1] >> v[3*i+2];
  t = new int[3*nt];
  for (int i = 0; i < nt; ++i) {
    istr >> k >> t[3*i] >> t[3*i+1] >> t[3*i+2];
    assert(k == 3);
  }
}

void readPolyhedronTRI (istream &istr, int &nv, double *&v, int &nt, int *&t)
{
  istr >> nv >> nt;
  v = new double [3*nv];
  for (int i = 0; i < nv; ++i)
    istr >> v[3*i] >> v[3*i+1] >> v[3*i+2];
  t = new int[3*nt];
  for (int i = 0; i < nt; ++i) {
    istr >> t[3*i] >> t[3*i+1] >> t[3*i+2];
  }
}

Polyhedron * polyhedron (int nv, double *v, int nt, int *t)
{
  Polyhedron *p = new Polyhedron;
  for (int i = 0; i < nv; ++i)
    p->addVertex(v[3*i], v[3*i+1], v[3*i+2]);
  for (int i = 0; i < nt; ++i)
    p->addFace(p->vertices[t[3*i]], p->vertices[t[3*i+1]], 
	       p->vertices[t[3*i+2]]);
  return p;
}

void BSPTree::find (const Point3 &p, BSPTrees &res)
{
  if (!left)
    res.push_back(this);
  else {
    int s = n.innerproduct(p).sign(false);
    if (s <= 0)
      left->find(p, res);
    if (s >= 0)
      right->find(p, res);
  }
}

void BSPTree::insert (PVertexCone *c)
{
  if (!left) {
    vertices.push_back(c->v);
    return;
  }
  int s = n.innerproduct(c->g[0]).sign(false);
  bool flag = s != 0;
  for (int i = 1; flag && i < c->g.size(); ++i)
    flag = s == n.innerproduct(c->g[i]).sign(false);
  if (!flag) {
    left->insert(c);
    right->insert(c);
  }
  else if (s == -1)
    left->insert(c);
  else
    right->insert(c);
}

void BSPTree::insert (PEdge *e, bool aflag)
{
  if (!left)
    if (aflag)
      aedges.push_back(e);
    else
      bedges.push_back(e);
  else {
    int s1 = n.innerproduct(e->face->n).sign(false), 
      s2 = n.innerproduct(e->twin->face->n).sign(false);
    if (s1 == -1 && s2 == -1)
      left->insert(e, aflag);
    else if (s1 == 1 && s2 == 1)
      right->insert(e, aflag);
    else {
      left->insert(e, aflag);
      right->insert(e, aflag);
    }
  }
}

void BSPTree::leafs (BSPTrees &res)
{
  if (left) {
    left->leafs(res);
    right->leafs(res);
  }
  else
    res.push_back(this);
}

BSPTree * BSPTreeBuild (PVertexCones &cones)
{
  Points3 pts;
  for (PVertexCones::iterator c = cones.begin(); c != cones.end(); ++c)
    pts.insert(pts.end(), (*c)->g.begin(), (*c)->g.end());
  BSPTree *tree = BSPTreeBuild(pts);
  for (PVertexCones::iterator c = cones.begin(); c != cones.end(); ++c)
    tree->insert(*c);
  return tree;
}

BSPTree * BSPTreeBuild (PEdges &aedges, PEdges &bedges)
{
  Points3 pts;
  for (PEdges::iterator e = aedges.begin(); e != aedges.end(); ++e) {
    pts.push_back((*e)->face->n);
    pts.push_back((*e)->twin->face->n);
  }
  for (PEdges::iterator e = bedges.begin(); e != bedges.end(); ++e) {
    pts.push_back((*e)->face->n);
    pts.push_back((*e)->twin->face->n);
  }
  BSPTree *tree = BSPTreeBuild(pts);
  for (PEdges::iterator e = aedges.begin(); e != aedges.end(); ++e)
    tree->insert(*e, true);
  for (PEdges::iterator e = bedges.begin(); e != bedges.end(); ++e)
    tree->insert(*e, false);
  return tree;
}

BSPTree * BSPTreeBuild (Points3 &pts, int nmax)
{
  if (pts.size() <= nmax)
    return new BSPTree;
  Point3 n = randomPoint3();
  Points3 pts1, pts2;
  for (Points3::iterator p = pts.begin(); p != pts.end(); ++p) {
    int s = n.innerproduct(*p).sign(false);
    if (s <= 0)
      pts1.push_back(*p);
    if (s >= 0)
      pts2.push_back(*p);
  }
  if (pts1.size() == pts.size() || pts2.size() == pts.size())
    return new BSPTree;
  BSPTree *left = BSPTreeBuild(pts1, nmax);
  BSPTree *right = BSPTreeBuild(pts2, nmax);
  return new BSPTree(n, left, right);
}

void CVertex::increasePrecision ()
{
  if (iedges[0] && iedges[1]) {
    iedges[0]->increasePrecision();
    iedges[1]->increasePrecision();
    face->increasePrecision();
    p = face->edgeIntersectionPoint(iedges[0], iedges[1]);
  }
  else if (iedges[0]) {
    iedges[0]->increasePrecision();
    face->increasePrecision();
    Point3 &t = iedges[0]->tail()->p, u = iedges[0]->u(), 
      &n = face->n, &a = face->boundary[0]->tail()->p;
    p = segmentPlaneIntersectionPoint(t, u, n, a);
  }
  else if (va && vb) {
    va->increasePrecision();
    vb->increasePrecision();
    p = va->p + vb->p;
  }
  else if (va) {
    va->increasePrecision();
    p = va->p;
  }
  else
    p.increasePrecision();
}

void CVertex::decreasePrecision ()
{
  if (iedges[0] && iedges[1]) {
    iedges[0]->decreasePrecision();
    iedges[1]->decreasePrecision();
    face->decreasePrecision();
  }
  else if (iedges[0]) {
    iedges[0]->decreasePrecision();
    face->decreasePrecision();
  }
  else if (va && vb) {
    va->decreasePrecision();
    vb->decreasePrecision();
  }
  else if (va)
    va->decreasePrecision();
  p.decreasePrecision();
}

void Triangle::increasePrecision ()
{
  a->increasePrecision();
  b->increasePrecision();
  c->increasePrecision();
}

void Triangle::decreasePrecision ()
{
  a->decreasePrecision();
  b->decreasePrecision();
  c->decreasePrecision();
}

void MetaCEdge::setBBox ()
{
  const Point3 &t = tail->p, &h = head->p;
  Parameter tx = t.getX(), ty = t.getY(), tz = t.getZ(),
    hx = h.getX(), hy = h.getY(), hz = h.getZ();
  bbox[0] = min(tx.lb(), hx.lb());
  bbox[1] = max(tx.ub(), hx.ub());
  bbox[2] = min(ty.lb(), hy.lb());
  bbox[3] = max(ty.ub(), hy.ub());
  bbox[4] = min(tz.lb(), hz.lb());
  bbox[5] = max(tz.ub(), hz.ub());
}

void MetaCEdge::addEdge (CEdge *e)
{
  if (ne < 2)
    edges[ne++] = e;
  else {
    if (ne == nmax) {
      CEdge **needges = new CEdge * [2*nmax-2];
      for (int i = 0; i + 2 < nmax; ++i)
	needges[i] = eedges[i];
      delete [] eedges;
      eedges = needges;
      nmax *= 2;
    }
    eedges[ne-2] = e;
    ++ne;
  }
}

CFace * MetaCEdge::intersectionFace (CVertex *v) const
{
  if (bflag)
    return v->face;
  for (int i = 0; i < 2; ++i)
    if (v->iedges[i]->edge != this)
      for (int j = 0; j < 2; ++j) {
	CFace *f = v->iedges[i]->edge->edges[j]->face;
	if (f != edges[0]->face && f != edges[1]->face)
	  return f;
      }
  return 0;
}

void MetaCEdge::formSubedges (bool allfaces, bool fail)
{
  if (allfaces)
    subedgesAll();
  else {
    prune1(fail);
    prune2();
    subedgesPruned();
  }
}

void MetaCEdge::subedgesAll () const
{
  for (int i = 0; i < ne; ++i) {
    CEdge *e = getEdge(i);
    for (int j = 0; j <= vertices.size(); ++j)
      e->addSubedge(j - 1);
  }
}

void MetaCEdge::prune1 (bool fail)
{
  int n = vertices.size() + 1;
  fflags = new bool [n];
  if (!manifold()) {
    for (int i = 0; i < n; ++i)
      fflags[i] = false;
    for (int i = 0; i + 1 < n; ++i)
      vertices[i]->fflag = false;
    return;
  }
  bool tflag = true;
  for (int i = 0; i + 1 < n; ++i) {
    CVertex *h = vertices[i];
    int sh = intersectionSign(h, fail);
    bool hflag = sh == -1;
    fflags[i] = tflag && hflag;
    tflag = !hflag;
  }
  fflags[n-1] = tflag;
  for (int i = 0; i + 1 < n; ++i)
    if (!fflags[i] && !fflags[i+1])
      vertices[i]->fflag = false;
}

bool MetaCEdge::manifold () const
{
  if (ne < 2)
    return false;
  bool flag = edges[0]->forward;
  if (edges[1]->forward != flag)
    return true;
  for (int i = 2; i < ne; ++i)
    if (eedges[i-2]->forward != flag)
      return true;
  return false;
}

int MetaCEdge::intersectionSign (CVertex *v, bool fail)
{ 
  CFace *f = intersectionFace(v);
  int s;
  try {
    s = f->n.innerproduct(u()).sign();
  }
  catch (SignException &exc) {
    if (fail)
      throw signException;
    f->increasePrecision();
    increasePrecision();
    s = f->n.innerproduct(u()).sign();
    f->decreasePrecision();
    decreasePrecision();
  }
  return s;
}

void MetaCEdge::prune2 () const
{
  for (int i = 0; i < vertices.size(); ++i)
    if (!fflags[i] && !fflags[i+1])
      vertices[i]->fflag = false;
}

void MetaCEdge::subedgesPruned () const
{
  int n = vertices.size() + 1;
  for (int i = 0; i < n; ++i)
    if (fflags[i] && (i == 0 || vertices[i-1]->fflag) &&
	(i + 1 == n || vertices[i]->fflag))
      for (int j = 0; j < ne; ++j)
	getEdge(j)->addSubedge(i - 1);
}

void MetaCEdge::increasePrecision ()
{
  tail->increasePrecision();
  head->increasePrecision();
}

void MetaCEdge::decreasePrecision ()
{
  tail->decreasePrecision();
  head->decreasePrecision();
}

bool EdgeVertexOrder::operator() (CVertex *v, CVertex *w) const
{
  if (v == w)
    return false;
  try {
    return e->u().innerproduct(w->p - v->p).sign() == 1;
  }
  catch (SignException &excp) {
    try {
      return orderAux(v, w);
    }
    catch (SignException &exc) {
      if (fail)
	throw signException;
      v->increasePrecision();
      w->increasePrecision();
      e->increasePrecision();
      bool res = e->u().innerproduct(w->p - v->p).sign() == 1;
      v->decreasePrecision();
      w->decreasePrecision();
      e->decreasePrecision();
      return res;    
    }
  }
}

bool EdgeVertexOrder::orderAux (CVertex *v, CVertex *w) const
{
  CFace *p = e->intersectionFace(v), *q = e->intersectionFace(w);
  Point3 &m = p->n, &n = q->n;
  CEdge *f = p->knownIntersection(q);
  if (f) {
    CVertex *c = f->tail()->iedges[0] ? f->tail() : f->head();
    CEdge *g = c->iedges[0];
    MetaCEdge *gg = g->edge;
    Point3 &t = e->tail->p, &h = e->head->p, eu = e->u(), fu = f->u();
    int s0 = c == f->tail() ? 1 : -1,
      s1 = (t - c->p).circulation(fu, eu).sign(),
      s2 = m.innerproduct(eu).sign(), s3 = n.innerproduct(eu).sign(), s4;
    if (c->face == q) {
      s4 = - n.innerproduct(g->u()).sign();
      for (int i = 0; i < gg->ne; ++i) {
	CEdge *h = gg->getEdge(i);
	if (h->face == p && h->forward == g->forward) {
	  s4 = - s4;
	  break;
	}
      }
    }
    else {
      s4 = m.innerproduct(g->u()).sign();
      for (int i = 0; i < gg->ne; ++i) {
	CEdge *h = gg->getEdge(i);
	if (h->face == q && h->forward == g->forward) {
	  s4 = - s4;
	  break;
	}
      }
    }
    return s0*s1*s2*s3*s4 == -1;
  }
  Point3 &a = p->boundary[0]->tail()->p, &b = q->boundary[0]->tail()->p;
  int s = m.innerproduct(q->boundary[0]->tail()->p - a).sign();
  bool flag = true;
  for (int i = 1; flag && i < q->nb; ++i) {
    CVertex *v = q->boundary[i]->tail();
    if (!p->boundaryVertex(v))
      flag = m.innerproduct(v->p - a).sign() == s;
  }
  if (flag) {
    int s1 = m.innerproduct(b - a).sign(), s2 = m.innerproduct(e->u()).sign();
    return s1*s2 == 1;
  }
  flag = true;
  s = n.innerproduct(p->boundary[0]->tail()->p - b).sign();
  for (int i = 1; flag && i < p->nb; ++i) {
    CVertex *v = p->boundary[i]->tail();
    if (!q->boundaryVertex(v))
      flag = n.innerproduct(v->p - b).sign() == s;
  }
  if (flag) {
    int s1 = n.innerproduct(b - a).sign(), s2 = n.innerproduct(e->u()).sign();
    return s1*s2 == 1;
  }
  throw signException;
}

CEdge::~CEdge ()
{ 
  for (Subedges::iterator s = subedges.begin(); s != subedges.end(); ++s) {
    delete (*s)->twin;
    delete *s;
  }
}

CFace * CEdge::otherFace () const
{
  return edge->edges[0] == this ? edge->edges[1]->face : edge->edges[0]->face;
}

bool CEdge::incident (CEdge *e) const
{
  return edge->tail == e->edge->tail || edge->head == e->edge->tail ||
    edge->tail == e->edge->head || edge->head == e->edge->head;
}

int CEdge::ccw (CFace *f, CFace *g)
{
  int s;
  try {
    s = u().circulation(f->n, g->n).sign();
  }
  catch (SignException &exc) {
    increasePrecision();
    f->increasePrecision();
    g->increasePrecision();
    s = u().circulation(f->n, g->n).sign();
    decreasePrecision();
    f->decreasePrecision();
    g->decreasePrecision();
  }
  return s;
}

void CEdge::addSubedge (int i)
{
  int nt = forward ? i : i + 1, nh = forward ? i + 1 : i;
  Subedge *e = new Subedge(nt, true, this, 0),
    *et = new Subedge(nh, false, this, e);
  e->twin = et;
  subedges.push_back(e);
}

Subedge * CEdge::findSubedge (CVertex *t, CVertex *h) const
{
  for (Subedges::const_iterator e = subedges.begin(); e != subedges.end(); ++e)
    if ((*e)->tail() == t && (*e)->head() == h)
      return *e;
  return 0;
}

CFace::~CFace ()
{
  for (CVertices::iterator v = vertices.begin(); v != vertices.end(); ++v)
    delete *v;
  for (int i = 0; i < ns; ++i)
    delete sfaces[i];
  delete [] sfaces;
}

void CFace::setBBox () 
{
  bbox[0] = bbox[2] = bbox[4] = 1e10;
  bbox[1] = bbox[3] = bbox[5] = -1e10;
  for (int i = 0; i < nb; ++i) {
    Point3 &p = boundary[i]->tail()->p;
    Parameter px = p.getX(), py = p.getY(), pz = p.getZ();
    if (px.lb() < bbox[0]) bbox[0] = px.lb();
    if (px.ub() > bbox[1]) bbox[1] = px.ub();
    if (py.lb() < bbox[2]) bbox[2] = py.lb();
    if (py.ub() > bbox[3]) bbox[3] = py.ub();
    if (pz.lb() < bbox[4]) bbox[4] = pz.lb();
    if (pz.ub() > bbox[5]) bbox[5] = pz.ub();
  }
}

bool CFace::coplanar (CFace *g) const
{
  return fv && fv == g->fv ||
    ffa && g->ffa && ffa->id == g->ffa->id ||
    fe && g->fe &&
    (boundary[0]->edge->avflag == g->boundary[0]->edge->avflag ? 
     fe->colinear(g->fe) && ff->colinear(g->ff) :
     fe->colinear(g->ff) && ff->colinear(g->fe));
}

bool CFace::canIntersect (CFace *g) const
{
  return !coplanar(g) &&
    (ffa && !fv ? !ffa->sharedEdge(g->ffa) : !sharedBoundaryEdge(g)) &&
    bboxOverlap(bbox, g->bbox);
}

bool CFace::sharedBoundaryEdge (CFace *g) const
{
  for (int i = 0; i < nb; ++i)
    for (int j = 0; j < g->nb; ++j)
      if (boundary[i]->edge == g->boundary[j]->edge)
	return true;
  return false;
}

bool CFace::intersection (CFace *g, bool justreport)
{
  CFace *ifaces[4];
  CEdge *iedges[4];
  CVertex *ivertices[2];
  bool flag = false;
  int ni = 0, nv = 0;
  if (intersection(g, ifaces, iedges, ni) && 
      g->intersection(this, ifaces, iedges, ni)) {
    for (int i = 0; i < ni; ++i) {
      Point3 p;
      if (ifaces[i]->intersection(iedges[i], p)) {
	if (justreport)
	  return true;
	flag = true;
	ivertices[nv++] = new CVertex(p, ifaces[i], iedges[i]);
      }
    }
    for (int i = 0; i < nv; ++i) {
      MetaCEdge *e = ivertices[i]->iedges[0]->edge;
      pthread_mutex_lock(&e->mutex);
      e->vertices.push_back(ivertices[i]);
      pthread_mutex_unlock(&e->mutex);
    }
  }
  return flag;
}

bool CFace::intersection (CFace *g, CFace **ifaces, CEdge **iedges, 
			  int &ni) const
{
  Point3 &c = g->boundary[0]->tail()->p;
  CVertex *t = boundary[nb-1]->tail();
  bool res = false, tbflag = !g->coplanarVertex(t), tsflag = false;
  if (tbflag)
    tsflag = g->n.innerproduct(t->p - c).sign() == 1;
  for (int i = 0; i < nb; ++i) {
    CVertex *h = boundary[i]->tail();
    bool hbflag = !g->coplanarVertex(h), hsflag = false;
    if (hbflag)
      hsflag = g->n.innerproduct(h->p - c).sign() == 1;
    if (tbflag && hbflag && tsflag != hsflag) {
      res = true;
      int ib = i == 0 ? nb - 1 : i - 1;
      if (g->canIntersect(boundary[ib])) {
	ifaces[ni] = g;
	iedges[ni] = boundary[ib];
	++ni;
      }
    }
    tbflag = hbflag;
    tsflag = hsflag;
  }
  return res;
}

bool CFace::coplanarVertex (CVertex *v) const
{
  if (boundaryVertex(v))
    return true;
  if (!v->va)
    return false;
  if (ffa) {
    if (!fv)
      return v->va->onFace(ffa);
    if (fv == v->va || fv == v->vb) {
      PVertex *w = fv == v->va ? v->vb : v->va;
      return w->onFace(ffa);
    }
    return false;
  }
  return boundary[0]->edge->avflag ? 
    fe->colinear(v->vb) && ff->colinear(v->va) :
    fe->colinear(v->va) && ff->colinear(v->vb);
}

bool CFace::canIntersect (CEdge *e) const
{
  MetaCEdge *ee = e->edge;
  if (e != ee->edges[0] || !bboxOverlap(ee->bbox, bbox))
    return false;
  if (ffa && !fv) {
    for (int i = 0; i < 3; ++i)
      if (ffa->boundary[i]->coplanar(e->edge->edge))
	return false;
  }
  else if (ffa || fe)
    for (int i = 0; i < nb; ++i) 
      if (boundary[i]->edge->edge->coplanar(e->edge->edge) &&
	  (boundary[i]->edge->avflag ? boundary[i]->tail()->va 
	   : boundary[i]->tail()->vb) ==
	  (e->edge->avflag ? e->tail()->va : e->tail()->vb))
	return false;
  return true;
}

bool CFace::intersection (CEdge *e, Point3 &p) const
{
  Point3 &a = boundary[0]->tail()->p, &t = e->tail()->p;
  p = segmentPlaneIntersectionPoint(t, e->u(), n, a);
  if (!bboxContains(bbox, p))
    return false;
  int coord = n.projectionCoordinate();
  Point2 q = p.project(coord);
  bool fflag = false;
  for (int i = 0; i < nb; ++i) {
    Point2 pc = boundary[i]->tail()->p.project(coord), 
      pv = boundary[i]->u().project(coord);
    int s = pv.crossproduct(q - pc).sign(false);
    if (s == -1)
      return false;
    if (s == 0)
      fflag = true;
  }
  if (fflag)
    throw signException;
  return true;
}

void CFace::formEdges (CFaces &gfaces, CVertices &vf, CVertices &vg)
{
  CVertex vd;
  vd.face = this;
  int i = 0, nv = vertices.size();
  while (i < nv) {
    CVertex *v = vertices[i];
    CFace *g = v->face;
    if (i + 1 < nv && vertices[i+1]->face == g) {
      gfaces.push_back(g);
      vf.push_back(v);
      vg.push_back(vertices[i+1]);
      i += 2;
    }
    else {
      bool eflag = false;
      for (int j = 0; !eflag && j < nb; ++j)
	if (g->boundaryVertex(boundary[j]->tail())) {
	  gfaces.push_back(g);
	  vf.push_back(v);
	  vg.push_back(boundary[j]->tail());
	  eflag = true;
	}
      if (!eflag && this < g) {
	CVertices::iterator w = lower_bound(g->vertices.begin(), g->vertices.end(),
					    &vd, VertexOrderF());
	if (w == g->vertices.end() || (*w)->face != this)
	  cerr << "missing edge/face intersection" << endl;
	gfaces.push_back(g);
	vf.push_back(v);
	vg.push_back(*w);
      }
      ++i;
    }
  }
}

MetaCEdge * CFace::formEdge (CFace *g, CVertex *v, CVertex *w)
{
  Point3 u = g->n.crossproduct(n);
  int s = u.innerproduct(w->p - v->p).sign();
  MetaCEdge *e = s == 1 ? new MetaCEdge(v, w, false, true, 0) : 
    new MetaCEdge(w, v, false, true, 0);
  CEdge *ef = new CEdge(e, true, this), *eg = new CEdge(e, false, g);
  pthread_mutex_lock(&mutex);
  edges.push_back(ef);
  pthread_mutex_unlock(&mutex);
  pthread_mutex_lock(&g->mutex);
  g->edges.push_back(eg);
  pthread_mutex_unlock(&g->mutex);
  return e;
}

void CFace::sortEdges ()
{
  vertices.clear();
  sort(edges.begin(), edges.end(), CFaceEdgeOrder());
}

void CFace::intersectEdges1 (CEdges &e1, CEdges &e2, CEdges &e3) const
{
  for (int j = 0; j < edges.size(); ++j) {
    CEdge *ej = edges[j];
    CFace *fj = ej->otherFace();
    if (this < fj)
      for (int k = j + 1; k < edges.size(); ++k) {
	CEdge *ek = edges[k];
	CFace *fk = ek->otherFace();
	if (this < fk && !ej->incident(ek) &&
	    bboxOverlap(ej->edge->bbox, ek->edge->bbox)) {
	  CEdge *el = fj->knownIntersection(fk);
	  if (el) {
	    e1.push_back(ej);
	    e2.push_back(ek);
	    e3.push_back(el);
	  }
	}
      }
  }
}

void CFace::intersectEdges2 (CEdge *ej, CEdge *ek, CEdge *el, CEdges &e1bad, 
			     CEdges &e2bad, CEdges &e3bad)
{
  try {
    if (intersection(ej, ek)) {
      Point3 p = edgeIntersectionPoint(ej, ek);
      CVertex *v = new CVertex(p, this, ej, ek);
      pthread_mutex_lock(&mutex);
      vertices.push_back(v);
      pthread_mutex_unlock(&mutex);
      pthread_mutex_lock(&ej->edge->mutex);
      ej->edge->vertices.push_back(v);
      pthread_mutex_unlock(&ej->edge->mutex);
      pthread_mutex_lock(&ek->edge->mutex);
      ek->edge->vertices.push_back(v);
      pthread_mutex_unlock(&ek->edge->mutex);
      pthread_mutex_lock(&el->edge->mutex);
      el->edge->vertices.push_back(v);
      pthread_mutex_unlock(&el->edge->mutex);
    }
  }
  catch (SignException &exc) {
    e1bad.push_back(ej);
    e2bad.push_back(ek);
    e3bad.push_back(el);
  }
}

bool CFace::intersection (CEdge *e, CEdge *f) const
{
  return intersectionAux(e, f) && intersectionAux(f, e);
}

bool CFace::intersectionAux (CEdge *e, CEdge *f) const
{
  if (e->otherFace()->coplanarVertex(f->tail()) ||
      e->otherFace()->coplanarVertex(f->head()))
    return false;
  int coord = n.projectionCoordinate();
  Point3 eu = e->u();
  int s1 = eu.crossproduct(e->tail()->p - f->tail()->p, coord).sign(),
    s2 = eu.crossproduct(e->tail()->p - f->head()->p, coord).sign();
  return s1 != s2;
}

Point3 CFace::edgeIntersectionPoint (CEdge *e, CEdge *f) const
{
  int coord = n.projectionCoordinate();
  Point3 p;
  try {
    p = lineIntersectionPoint(e->tail()->p, e->u(), f->tail()->p, f->u(), coord);
  }
  catch (SignException &exc) {
    e->increasePrecision();
    f->increasePrecision();
    p = lineIntersectionPoint(e->tail()->p, e->u(), f->tail()->p, f->u(), coord);
    e->decreasePrecision();
    f->decreasePrecision();
  }
  return p;
}

bool CFace::boundaryVertex (PVertex *v) const
{
  for (int i = 0; i < nb; ++i)
    if (boundary[i]->tail()->va == v)
      return true;
  return false;
}

bool CFace::boundaryVertex (CVertex *v) const
{
  for (int i = 0; i < nb; ++i)
    if (boundary[i]->tail() == v)
      return true;
  return false;
}

CEdge * CFace::knownIntersection (CFace *g)
{
  if (!canIntersect(g))
    return 0;
  MetaCEdge m;
  CEdge d(&m, true, this), dt(&m, false, g);
  CEdges::iterator iter = lower_bound(edges.begin(), edges.end(), &d, 
				       CFaceEdgeOrder());
  if (iter != edges.end() && (*iter)->otherFace() == g)
    return *iter;
  return 0;
}

void CFace::sortVertices ()
{
  for (int i = 0; i < nb; ++i) {
    CVertices &vi = boundary[i]->edge->vertices;
    vertices.insert(vertices.end(), vi.begin(), vi.end());
  }
  sort(vertices.begin(), vertices.end(), VertexOrderF());
}

void CFace::subdivide (bool allfaces, bool fail)
{
  SubedgeMap smap;
  subdivideInit(fail, smap);
  Subedges outer, inner;
  for (SubedgeMap::iterator i = smap.begin(); i != smap.end(); ++i) {
    for (Subedges::iterator e = i->second->begin(); e != i->second->end(); ++e)
    if (!(*e)->traversed) {
      bool oflag;
      if (loop(*e, allfaces, fail, oflag))
	if (oflag)
	  outer.push_back(*e);
	else
	  inner.push_back(*e);
    }
    delete i->second;
  }
  ns = outer.size();
  sfaces = new SFace * [ns];
  for (int i = 0; i < ns; ++i) {
    sfaces[i] = new SFace(this);
    sfaces[i]->addBoundary(outer[i]);
  }
  for (Subedges::iterator e = inner.begin(); e != inner.end(); ++e) {
    SFace *f = enclosingFace((*e)->tail(), fail);
    if (f)
      f->addBoundary(*e);
  }
}

void CFace::subdivideInit (bool fail, SubedgeMap &smap) const
{
  for (int i = 0; i < nb; ++i)
    subdivideInitEdge(smap, boundary[i], true);
  for (CEdges::const_iterator e = edges.begin(); e != edges.end(); ++e)
    subdivideInitEdge(smap, *e, false);
  for (SubedgeMap::iterator i = smap.begin(); i != smap.end(); ++i)
    subdivideInitVertex(fail, *i->second);
}

void CFace::subdivideInitEdge (SubedgeMap &smap, CEdge *e, bool tflag) const
{
  for (Subedges::iterator s = e->subedges.begin(); s != e->subedges.end(); ++s) 
    for (int i = 0; i < 2; ++i) {
      Subedge *si = i == 0 ? *s : (*s)->twin;
      if (tflag && i == 1)
	si->traversed = true;
      CVertex *v = si->tail();
      SubedgeMap::iterator iter = smap.find(v);
      Subedges *sedges = 0;
      if (iter == smap.end()) {
	sedges = new Subedges;
	smap.insert(pair<CVertex *, Subedges *>(v, sedges));
      }
      else
	sedges = iter->second;
      sedges->push_back(si);
    }
}

void CFace::subdivideInitVertex (bool fail, Subedges &subedges) const
{
  if (subedges.size() > 2)
    sort(subedges.begin(), subedges.end(), SubedgeOrder(this, fail));
  Subedge *a = *(subedges.end()-1);
  for (Subedges::iterator b = subedges.begin(); b != subedges.end(); ++b) {
    a->next = *b;
    a = *b;
  }
}

bool CFace::loop (Subedge *e, bool allfaces, bool fail, bool &outer) const
{
  int coord = n.projectionCoordinate();
  bool res = true, manifold = true;
  outer = false;
  Subedge *f = e;
  do {
    f->traversed = true;
    res = res && f->forward;
    outer = outer || f->edge->edge->bflag;
    Subedge *nf = f->twin->next;
    if (nf == f->twin)
      manifold = false;
    f = nf;
  }
  while (f != e);
  if (!manifold || !allfaces && !res)
    return false;
  if (outer)
    return true;
  Subedge *g = e;
  f = e;
  do {
    if (f->head() != g->head()) {
      bool flag;
      try {
	flag = f->head()->p.project(coord).getX() <
	  g->head()->p.project(coord).getX();
      }
      catch (SignException &exc) {
	if (fail)
	  throw signException;
	f->head()->increasePrecision();
	g->head()->increasePrecision();
	flag = f->head()->p.project(coord).getX() <
	  g->head()->p.project(coord).getX();
	f->head()->decreasePrecision();
	g->head()->decreasePrecision();
      }
      if (flag)
	g = f;
    }
    f = f->twin->next;
  }
  while (f != e);
  Subedge *ng = g->twin->next;
  int s;
  try {
    Point2 gu = g->u().project(coord), ngu = ng->u().project(coord);
    s = gu.crossproduct(ngu).sign();
  }
  catch (SignException &exc) {
    if (fail)
      throw signException;
    g->edge->edge->increasePrecision();
    ng->edge->edge->increasePrecision();
    Point2 gu = g->u().project(coord), ngu = ng->u().project(coord);
    s = gu.crossproduct(ngu).sign();
    g->edge->edge->decreasePrecision();
    ng->edge->edge->decreasePrecision();
  }
  outer = s == 1;
  return true;
}

SFace * CFace::enclosingFace (CVertex *v, bool fail) const
{
  int coord = n.projectionCoordinate();
  Parameter k, ki;
  SFace *res = 0;
  for (int i = 0; i < ns; ++i)
    if (sfaces[i]->contains(v, coord, fail, ki)) {
      if (!res || ki < k) {
	k = ki;
	res = sfaces[i];
      }
    }
  return res;
}

void CFace::increasePrecision ()
{
  for (int i = 0; i < nb; ++i)
    boundary[i]->tail()->increasePrecision();
  n = boundary[0]->u().crossproduct(boundary[1]->u());
}

void CFace::decreasePrecision ()
{
  for (int i = 0; i < nb; ++i)
    boundary[i]->tail()->decreasePrecision();
  n.decreasePrecision();
}

void CFace::decreasePrecisionBV () const
{
  for (int i = 0; i < nb; ++i)
    for (int j = 0; j < boundary[i]->edge->vertices.size(); ++j)
      boundary[i]->edge->vertices[j]->decreasePrecision();
}

Convolution::~Convolution () 
{
  for (CFaces::iterator f = faces.begin(); f != faces.end(); ++f)
    delete *f;
  for (MetaCEdges::iterator e = edges.begin(); e != edges.end(); ++e) {
    if ((*e)->bflag)
      for (CVertices::iterator v = (*e)->vertices.begin();
	   v != (*e)->vertices.end(); ++v)
	delete *v;
    for (int i = 0; i < (*e)->ne; ++i)
      delete (*e)->getEdge(i);
    delete *e;
  }
  for (CVertices::iterator v = vertices.begin(); v !=vertices.end(); ++v)
    delete *v;
}

void Convolution::addFace (CFace *f)
{
  f->id = ++IdCounter;
  faces.push_back(f);
}

void Convolution::sum (const Polyhedron *a, const Polyhedron *b)
{
  VVMap vvmap;
  MetaCEdgeSet eset;
  sumVF(a, b, true, vvmap, eset);
  sumVF(b, a, false, vvmap, eset);
  sumEE(a, b, vvmap, eset);
}

void Convolution::sumVF (const Polyhedron *a, const Polyhedron *b, 
			 bool avflag, VVMap &vvmap, MetaCEdgeSet &eset)
{
  PVertexCones convex;
  for (PVertices::const_iterator v = a->vertices.begin(); 
       v != a->vertices.end(); ++v) {
    PVertexCone *c;
    try {
      c = convexCone(*v);
    }
    catch (SignException &exc) {
      (*v)->increasePrecisionCone();
      c = convexCone(*v);
      (*v)->decreasePrecisionCone();
      if (c)
	c->decreasePrecision();
    }
    if (c)
      convex.push_back(c);
  }
  BSPTree *tree = BSPTreeBuild(convex);
  for (PVertexCones::iterator c = convex.begin(); c != convex.end(); ++c)
    delete *c;
  for (PFaces::const_iterator f = b->faces.begin(); f != b->faces.end(); ++f) {
    BSPTrees leafs;
    tree->find((*f)->n, leafs);
    PVertexset vset;
    for (BSPTrees::iterator l = leafs.begin(); l != leafs.end(); ++l)
      for (PVertices::iterator v = (*l)->vertices.begin(); 
	   v != (*l)->vertices.end(); ++v)
	vset.insert(*v);
    for (PVertexset::iterator v = vset.begin(); v != vset.end(); ++v) {
      bool flag;
      try {
	flag = (*v)->compatible(*f);
      }
      catch (SignException &exc) {
	(*v)->increasePrecisionCone();
	(*f)->increasePrecision();
	flag = (*v)->compatible(*f);
	(*v)->decreasePrecisionCone();
	(*f)->decreasePrecision();
      }
      if (flag)
	sumVF(*v, *f, avflag, vvmap, eset);
    }
  }
  delete tree;
}

void Convolution::sumVF (PVertex *v, PFace *f, bool avflag, VVMap &vvmap,
			 MetaCEdgeSet &eset)
{
  CFace *face = new CFace(v, f);
  addFace(face);
  for (int i = 0; i < 3; ++i)
    face->boundary[i] = sumVE(v, f->boundary[i], avflag, vvmap, eset, face);
  face->n = face->boundary[0]->u().crossproduct(face->boundary[1]->u());
  face->setBBox();
}

void Convolution::sumEE (const Polyhedron *a, const Polyhedron *b, VVMap &vvmap,
			 MetaCEdgeSet &eset)
{
  PEdges aedges, bedges;
  a->convexEdges(aedges);
  b->convexEdges(bedges);
  BSPTree *tree = BSPTreeBuild(aedges, bedges);
  BSPTrees leafs;
  tree->leafs(leafs);
  set<EEPair> eeset;
  for (BSPTrees::iterator l = leafs.begin(); l != leafs.end(); ++l)
    for (PEdges::iterator e = (*l)->aedges.begin(); e != (*l)->aedges.end(); ++e)
      for (PEdges::iterator f = (*l)->bedges.begin(); f != (*l)->bedges.end(); ++f)
	if (eeset.find(EEPair(*e, *f)) == eeset.end()) {
	  bool aflag;
	  if ((*e)->compatible(*f, aflag))
	    sumEE(*e, *f, aflag, vvmap, eset);
	  eeset.insert(EEPair(*e, *f));
	}
  delete tree;
}

void Convolution::sumEE (PEdge *e, PEdge *f, bool aflag, VVMap &vvmap, 
			 MetaCEdgeSet &eset)
{
  PEdge *ee = aflag ? e : f, *ff = aflag ? f : e;
  CFace *face = new CFace(ee, ff);
  addFace(face);
  face->boundary[0] = sumVE(ff->tail, ee, !aflag, vvmap, eset, face);
  face->boundary[1] = sumVE(ee->head(), ff, aflag, vvmap, eset, face);
  face->boundary[2] = sumVE(ff->head(), ee->twin, !aflag, vvmap, eset, face);
  face->boundary[3] = sumVE(ee->tail, ff->twin, aflag, vvmap, eset, face);
  face->n = face->boundary[0]->u().crossproduct(face->boundary[1]->u());
  face->setBBox();
}

CEdge * Convolution::sumVE (PVertex *v, PEdge *e, bool avflag, VVMap &vvmap,
			    MetaCEdgeSet &eset, CFace *face)
{
  CVertex *t = sumVV(v, e->tail, avflag, vvmap),
    *h = sumVV(v, e->head(), avflag, vvmap);
  CEdge *f = addEdge(t, h, face, avflag, e, eset);
  return f;
}

CVertex * Convolution::sumVV (PVertex *a, PVertex *b, bool flag, VVMap &vvmap)
{
  PVertex *aa = flag ? a : b, *bb = flag ? b : a;
  VVPair vp(aa, bb);
  VVMap::iterator iter = vvmap.find(vp);
  if (iter != vvmap.end())
    return iter->second;
  CVertex *v = new CVertex(aa->p + bb->p, aa, bb);
  vertices.push_back(v);
  vvmap.insert(pair<VVPair, CVertex *>(vp, v));
  return v;
}

void Convolution::subdivide (bool allfaces, bool rbflag)
{
  intersectFaces(rbflag);
  sortVerticesF();
  formEdges();
  sortEdges();
  intersectEdges();
  sortVertices();
  formSubedges(allfaces);
  subdivideFaces(allfaces);
}

void Convolution::intersectFaces (bool rbflag) const
{
  CFaces fa, ga;
  KDTreeBuild(rbflag, fa, ga);
  int counter = 0, step = 1000;
  ThreadData td[NPROC];
  for (int i = 0; i < NPROC; ++i) {
    td[i].s = &counter;
    td[i].d = step;
    td[i].n = fa.size();
    td[i].fa = &fa;
    td[i].ga = &ga;
  }
  runThreaded(td, IntersectFaces);
  fa.clear();
  ga.clear();
  for (int i = 0; i < NPROC; ++i) {
    fa.insert(fa.end(), td[i].fbad.begin(), td[i].fbad.end());
    ga.insert(ga.end(), td[i].gbad.begin(), td[i].gbad.end());
    td[i].fbad.clear();
    td[i].gbad.clear();
  }
  if (fa.empty())
    return;
  for (int i = 0; i < NPROC; ++i)
    td[i].n = fa.size();
  for (int i = 0; i < fa.size(); ++i) {
    fa[i]->increasePrecision();
    ga[i]->increasePrecision();
  }
  runThreaded(td, IntersectFaces);
  for (int i = 0; i < NPROC; ++i)
    if (!(td[i].fbad.empty() && td[i].gbad.empty()))
      throw signException;
  for (int i = 0; i < fa.size(); ++i) {
    fa[i]->decreasePrecision();
    fa[i]->decreasePrecisionBV();
    ga[i]->decreasePrecision();
    ga[i]->decreasePrecisionBV();
  }
}

void Convolution::KDTreeBuild (bool rbflag, CFaces &fa, CFaces &ga) const
{
  CVertices xpts;
  for (CVertices::const_iterator v = vertices.begin(); v != vertices.end(); ++v)
    xpts.push_back(*v);
  CVertices ypts(xpts), zpts(xpts);
  sort(xpts.begin(), xpts.end(), AxisOrder(1));
  sort(ypts.begin(), ypts.end(), AxisOrder(2));
  sort(zpts.begin(), zpts.end(), AxisOrder(3));
  FNodes nodes;
  for (CFaces::const_iterator f = faces.begin(); f != faces.end(); ++f)
    nodes.push_back(FNode(*f));
  ::KDTreeBuild(xpts, ypts, zpts, nodes, 10, 0, rbflag, fa, ga);  
}

void Convolution::sortVerticesF () const
{
  int counter = 0, step = 1000;
  ThreadData td[NPROC];
  for (int i = 0; i < NPROC; ++i) {
    td[i].s = &counter;
    td[i].d = step;
    td[i].n = faces.size();
    td[i].fa = &faces;
  }
  runThreaded(td, SortVerticesF);
}

void Convolution::formEdges ()
{
  int counter = 0, step = 1000;
  ThreadData td[NPROC];
  for (int i = 0; i < NPROC; ++i) {
    td[i].s = &counter;
    td[i].d = step;
    td[i].n = faces.size();
    td[i].fa = &faces;
    td[i].edges = &edges;
  }
  runThreaded(td, FormEdges);
  for (int i = 0; i < NPROC; ++i) {
    edges.insert(edges.end(), td[i].ebad.begin(), td[i].ebad.end());
    for (int j = 0; j < td[i].fbad.size(); ++j) {
      CFace *f = td[i].fbad[j], *g = td[i].gbad[j];
      CVertex *v = td[i].vbad[j], *w = td[i].wbad[j];
      f->increasePrecision();
      g->increasePrecision();
      v->increasePrecision();
      w->increasePrecision();
      edges.push_back(f->formEdge(g, v, w));
      f->decreasePrecision();
      g->decreasePrecision();
      v->decreasePrecision();
      w->decreasePrecision();
    }
  }
}

void Convolution::sortEdges () const
{
  int counter = 0, step = 1000;
  ThreadData td[NPROC];
  for (int i = 0; i < NPROC; ++i) {
    td[i].s = &counter;
    td[i].d = step;
    td[i].n = faces.size();
    td[i].fa = &faces;
  }
  runThreaded(td, SortEdges);
}

void Convolution::intersectEdges () const
{
  int counter = 0, step = 1000;
  ThreadData td[NPROC];
  for (int i = 0; i < NPROC; ++i) {
    td[i].s = &counter;
    td[i].d = step;
    td[i].n = faces.size();
    td[i].fa = &faces;
  }
  runThreaded(td, IntersectEdges1);
  CEdges e1, e2, e3;
  for (int i = 0; i < NPROC; ++i) {
    e1.insert(e1.end(), td[i].e1bad.begin(), td[i].e1bad.end());
    e2.insert(e2.end(), td[i].e2bad.begin(), td[i].e2bad.end());
    e3.insert(e3.end(), td[i].e3bad.begin(), td[i].e3bad.end());
    td[i].e1bad.clear();
    td[i].e2bad.clear();
    td[i].e3bad.clear();
  }
  for (int i = 0; i < NPROC; ++i) {
    td[i].n = e1.size();
    td[i].e1 = &e1;
    td[i].e2 = &e2;
    td[i].e3 = &e3;
  }
  runThreaded(td, IntersectEdges2);
  e1.clear();
  e2.clear();
  e3.clear();
  for (int i = 0; i < NPROC; ++i) {
    e1.insert(e1.end(), td[i].e1bad.begin(), td[i].e1bad.end());
    e2.insert(e2.end(), td[i].e2bad.begin(), td[i].e2bad.end());
    e3.insert(e3.end(), td[i].e3bad.begin(), td[i].e3bad.end());
    td[i].e1bad.clear();
    td[i].e2bad.clear();
    td[i].e3bad.clear();
  }
  if (e1.empty())
    return;
  for (int i = 0; i < NPROC; ++i) {
    td[i].n = e1.size();
    td[i].e1 = &e1;
    td[i].e2 = &e2;
    td[i].e3 = &e3;
  }
  for (int i = 0; i < e1.size(); ++i) {
    e1[i]->increasePrecision();
    e2[i]->increasePrecision();
  }
  runThreaded(td, IntersectEdges2);
  for (int i = 0; i < e1.size(); ++i) {
    e1[i]->decreasePrecision();
    for (CVertices::iterator v = e1[i]->edge->vertices.begin();
	 v != e1[i]->edge->vertices.end(); ++v)
      (*v)->decreasePrecision();
    e2[i]->decreasePrecision();
  }
  for (int i = 0; i < NPROC; ++i)
    if (!(td[i].e1bad.empty()))
      throw signException;
}

void Convolution::sortVertices ()
{
  int counter = 0, step = 1000;
  ThreadData td[NPROC];
  for (int i = 0; i < NPROC; ++i) {
    td[i].s = &counter;
    td[i].d = step;
    td[i].n = edges.size();
    td[i].edges = &edges;
  }
  runThreaded(td, SortVertices);
  MetaCEdges ebad;
  for (int i = 0; i < NPROC; ++i) {
    ebad.insert(ebad.end(), td[i].ebad.begin(), td[i].ebad.end());
    td[i].ebad.clear();
  }
  if (ebad.empty())
    return;
  for (MetaCEdges::iterator e = ebad.begin(); e != ebad.end(); ++e)
    sort((*e)->vertices.begin(), (*e)->vertices.end(), 
	 EdgeVertexOrder(*e, false));
}

void Convolution::formSubedges (bool allfaces)
{
  int counter = 0, step = 1000;
  ThreadData td[NPROC];
  for (int i = 0; i < NPROC; ++i) {
    td[i].s = &counter;
    td[i].d = step;
    td[i].n = edges.size();
    td[i].edges = &edges;
    td[i].allfaces = allfaces;
  }
  runThreaded(td, FormSubedges);
  MetaCEdges ebad;
  for (int i = 0; i < NPROC; ++i) {
    ebad.insert(ebad.end(), td[i].ebad.begin(), td[i].ebad.end());
    td[i].ebad.clear();
  }
  for (MetaCEdges::iterator e = ebad.begin(); e != ebad.end(); ++e)
    (*e)->formSubedges(allfaces, false);
}

void Convolution::subdivideFaces (bool allfaces) const
{
  int counter = 0, step = 1000;
  ThreadData td[NPROC];
  for (int i = 0; i < NPROC; ++i) {
    td[i].s = &counter;
    td[i].d = step;
    td[i].n = faces.size();
    td[i].fa = &faces;
    td[i].allfaces = allfaces;
  }
  runThreaded(td, SubdivideFaces);
  CFaces fbad;
  for (int i = 0; i < NPROC; ++i)
    fbad.insert(fbad.end(), td[i].fbad.begin(), td[i].fbad.end());
  for (CFaces::iterator f = fbad.begin(); f != fbad.end(); ++f)
    (*f)->subdivide(allfaces, false);
}

void Convolution::merge (const Polyhedron *p, bool aflag, bool perturb)
{
  PVCVMap vmap;
  CVertex * v[3];
  CEdge * b[3];
  MetaCEdgeSet eset;
  for (PFaces::const_iterator f = p->faces.begin(); f != p->faces.end(); ++f) {
    for (int i = 0; i < 3; ++i)
      v[i] = getVertex((*f)->boundary[i]->tail, vmap, perturb);
    CFace *face = new CFace(aflag);
    if (!perturb)
      face->ffa = *f;
    addFace(face);
    for (int i = 0; i < 3; ++i)
      face->boundary[i] = addEdge(v[i], v[(i+1)%3], face, true,
				  perturb ? 0 : (*f)->boundary[i], eset);
    face->n = face->boundary[0]->u().crossproduct(face->boundary[1]->u());
    face->setBBox();
  }
}

CVertex * Convolution::getVertex (PVertex *v, PVCVMap &vmap, bool perturb)
{
  PVCVMap::iterator iter = vmap.find(v);
  if (iter != vmap.end())
    return iter->second;
  CVertex *w;
  if (perturb)
    w = new CVertex(v->p.perturb());
  else {
    w = new CVertex(v->p);
    w->va = v;
  }
  vertices.push_back(w);
  vmap.insert(pair<PVertex *, CVertex *>(v, w));
  return w;
}

CEdge * Convolution::addEdge (CVertex *t, CVertex *h, CFace *face,
			      bool avflag, PEdge *e, MetaCEdgeSet &eset)
{
  CVertex *tt = t < h ? t : h, *hh = t < h ? h : t;
  MetaCEdge *f = new MetaCEdge(tt, hh, true, avflag, e);
  MetaCEdgeSet::iterator iter = eset.find(f);
  if (iter == eset.end()) {
    eset.insert(f);
    edges.push_back(f);
  }
  else {
    delete f;
    f = *iter;
  }
  return new CEdge(f, t == tt, face);
}

void Convolution::formShells (bool allfaces, Shells &shells) const
{
  for (CFaces::const_iterator f = faces.begin(); f != faces.end(); ++f)
    for (int i = 0; i < (*f)->ns; ++i)
      for (int j = 0; j < 2; ++j) {
	HSFace &h = (*f)->sfaces[i]->hfaces[j];
	if (!(h.flag || j == 1 && !allfaces)) {
	  Shell *s = h.formShell(allfaces);
	  if (s)
	    shells.push_back(s);
	}
      }
  sort(shells.begin(), shells.end(), ShellOrder());
}

bool Convolution::testIntersection (PVertex *at, PVertex *bt) const
{
  CFaces fa, ga;
  KDTreeBuild(true, fa, ga);
  for (int i = 0; i < fa.size(); ++i) {
    CFace *f = fa[i], *g = ga[i];
    try {
      if (!f->boundaryVertex(at) && !g->boundaryVertex(bt) &&
	  f->intersection(g, true))
	return true;
    }
    catch (SignException &exc) {
      f->increasePrecision();
      g->increasePrecision();
      bool res = f->intersection(g, true);
      f->decreasePrecision();
      g->decreasePrecision();
      if (res)
	return true;
    }
  }
  return false;
}

void KDTreeBuild (CVertices &pts1, CVertices &pts2, CVertices &pts3,
		  const FNodes &nodes, int nmax, int coord, bool rbflag,
		  CFaces &fa, CFaces &ga)
{
  int n = pts1.size();
  if (n <= nmax || coord == 63) {
    KDTreeLeaf(nodes, rbflag, fa, ga);
    return;
  }
  int mcoord = 1 + coord % 3;
  CVertex *pm = pts1[n/2];
  CVertices pts1a, pts1b, pts2a, pts2b, pts3a, pts3b;
  partition(pts1, pm, mcoord, pts1a, pts1b);
  partition(pts2, pm, mcoord, pts2a, pts2b);
  partition(pts3, pm, mcoord, pts3a, pts3b);
  if (pts1a.size() == n || pts1b.size() == n) {
    KDTreeLeaf(nodes, rbflag, fa, ga);
    return;
  }
  FNodes nodes1, nodes2;
  partition(nodes, pm, coord, nodes1, nodes2);
  KDTreeBuild(pts2a, pts3a, pts1a, nodes1, nmax, coord + 1, rbflag, fa, ga);
  KDTreeBuild(pts2b, pts3b, pts1b, nodes2, nmax, coord + 1, rbflag, fa, ga);
}

void KDTreeLeaf (const FNodes &nodes, bool rbflag, CFaces &fa, CFaces &ga)
{
  for (FNodes::const_iterator a = nodes.begin(); a != nodes.end(); ++a) {
    FNodes::const_iterator b = a;
    ++b;
    while (b != nodes.end()) {
      if (a->compatible(*b) && 
	  !(rbflag && a->face->flag == b->face->flag) &&
	  a->face->canIntersect(b->face)) {
	fa.push_back(a->face);
	ga.push_back(b->face);
      }
      ++b;
    }
  }
}

void partition (const CVertices &pts, CVertex *p, int coord,
		CVertices &pts1, CVertices &pts2)
{
  Parameter c = p->p.coordinate(coord);
  for (int i = 0; i < pts.size(); ++i) {
    Parameter pc = pts[i]->p.coordinate(coord);
    if (c.ub() < pc.lb())
      pts2.push_back(pts[i]);
    else
      pts1.push_back(pts[i]);
  }
}

void partition (const FNodes &nodes, CVertex *p, int coord,
		FNodes &nodes1, FNodes &nodes2)
{
  int mcoord = 1 + coord % 3;
  Parameter c = p->p.coordinate(mcoord);
  int j = 2*(mcoord - 1);
  for (FNodes::const_iterator a = nodes.begin(); a != nodes.end(); ++a) {
    double *bbox = a->face->bbox;
    if (bbox[j+1] < c.lb())
      nodes1.push_back(*a);
    else if (c.ub() < bbox[j])
      nodes2.push_back(*a);
    else {
      nodes1.push_back(*a);
      FNode b(*a);
      b.label += 1u << coord;
      nodes2.push_back(b);
    }
  }
}

bool bboxOverlap (const double *a, const double *b)
{
  for (int i = 0; i < 3; ++i)
    if (a[2*i+1] < b[2*i] || b[2*i+1] < a[2*i])
      return false;
  return true;
}

void runThreaded (ThreadData *td, ThreadType type)
{
  double t0 = getTime();
  for (int i = 0; i < NPROC; ++i) {
    *td[i].s = 0;
    td[i].type = type;
  }
  pthread_t threads[NPROC];
  for (int i = 0; i < NPROC; ++i)
    pthread_create(threads + i, 0, runThread, (void *) (td + i));
  for (int i = 0; i < NPROC; ++i)
    pthread_join(threads[i], 0);
  double t1 = getTime() - t0;
  //cerr << "time " << type << " " << t1 << endl;
}

pthread_mutex_t MutexRunThread = PTHREAD_MUTEX_INITIALIZER;

void * runThread (void *ptr)
{
  ThreadData *td = (ThreadData *) ptr;
  int emax = td->n;
  while (*td->s < emax) {
    int s;
    pthread_mutex_lock(&MutexRunThread);
    s = *td->s;
    *td->s += td->d;
    pthread_mutex_unlock(&MutexRunThread);
    int e = min(s + td->d, emax);
    switch (td->type) {
    case IntersectFaces:
      for (int i = s; i < e; ++i) 
	intersectFaces(td->fa->at(i), td->ga->at(i), td->fbad, td->gbad);
      break;
    case SortVerticesF:
      for (int i = s; i < e; ++i) 
	td->fa->at(i)->sortVertices();
      break;
    case FormEdges:
      for (int i = s; i < e; ++i)
	formEdges(td->fa->at(i), td->ebad, td->fbad, td->gbad, 
		  td->vbad, td->wbad);
      break;
    case SortEdges:
      for (int i = s; i < e; ++i) 
	td->fa->at(i)->sortEdges();
      break;
    case IntersectEdges1:
      for (int i = s; i < e; ++i) 
	td->fa->at(i)->intersectEdges1(td->e1bad, td->e2bad, td->e3bad);
      break;
    case IntersectEdges2:
      for (int i = s; i < e; ++i) {
	CEdge *ei = td->e1->at(i), *ej = td->e2->at(i), *ek = td->e3->at(i);
	CFace *f = ei->face;
	f->intersectEdges2(ei, ej, ek, td->e1bad, td->e2bad, td->e3bad);
      }
      break;
    case SortVertices:
      for (int i = s; i < e; ++i) 
	sortVertices(td->edges->at(i), td->ebad);
      break;
    case FormSubedges:
      for (int i = s; i < e; ++i) 
	formSubedges(td->edges->at(i), td->allfaces, td->ebad);
      break;
    case SubdivideFaces:
      for (int i = s; i < e; ++i)
	subdivideFaces(td->fa->at(i), td->allfaces, td->fbad);
      break;
    }
  }
}

void intersectFaces (CFace *f, CFace *g, CFaces &fbad, CFaces &gbad)
{
  try {
    f->intersection(g);
  }
  catch (SignException &exc) {
    fbad.push_back(f);
    gbad.push_back(g);
  }
}

void formEdges (CFace *f, MetaCEdges &edges, CFaces &fbad, CFaces &gbad,
		CVertices &vbad, CVertices &wbad)
{
  CFaces gfaces;
  CVertices vf, vg;
  f->formEdges(gfaces, vf, vg);
  for (int i = 0; i < gfaces.size(); ++i)
    try {
      MetaCEdge *e = f->formEdge(gfaces[i], vf[i], vg[i]);
      edges.push_back(e);
    }
    catch (SignException &exc) {
      fbad.push_back(f);
      gbad.push_back(gfaces[i]);
      vbad.push_back(vf[i]);
      wbad.push_back(vg[i]);
    }
}

void sortVertices (MetaCEdge *e, MetaCEdges &ebad)
{
  CVertices ev = e->vertices;
  try {
    sort(e->vertices.begin(), e->vertices.end(), EdgeVertexOrder(e, true));
  }
  catch (SignException &exc) {
    e->vertices = ev;
    ebad.push_back(e);
  }
}

void formSubedges (MetaCEdge *e, bool allfaces, MetaCEdges &ebad)
{
  try {
    e->formSubedges(allfaces, true);
  }
  catch (SignException &exc) {
    ebad.push_back(e);
  }
}  

void subdivideFaces (CFace *f, bool allfaces, CFaces &fbad)
{
  try {
    f->subdivide(allfaces, true);
  }
  catch (SignException &exc) {
    if (f->ns) {
      for (int i = 0; i < f->ns; ++i)
	delete f->sfaces[i];
      delete [] f->sfaces;
      f->sfaces = 0;
      f->ns = 0;
    }
    for (int i = 0; i < f->nb; ++i)
      for (Subedges::iterator s = f->boundary[i]->subedges.begin();
	   s != f->boundary[i]->subedges.end(); ++s) {
	(*s)->traversed = (*s)->twin->traversed = false;
	(*s)->next = (*s)->twin->next = 0;
      }
    for (CEdges::iterator e = f->edges.begin(); e != f->edges.end(); ++e)
      for (Subedges::iterator s = (*e)->subedges.begin();
	   s != (*e)->subedges.end(); ++s) {
	(*s)->traversed = (*s)->twin->traversed = false;
	(*s)->next = (*s)->twin->next = 0;
      }
    fbad.push_back(f);
  }
}

CVertex * Subedge::tail () const
{
  MetaCEdge *e = edge->edge;
  if (nt == -1)
    return e->tail;
  if (nt == e->vertices.size())
    return e->head;
  return e->vertices[nt];
}
 
CVertex * Subedge::head () const
{
  MetaCEdge *e = edge->edge;
  int nh = forward == edge->forward ? nt + 1 : nt - 1;
  if (nh == -1)
    return e->tail;
  if (nh == e->vertices.size())
    return e->head;
  return e->vertices[nh];
}

bool SubedgeOrder::operator() (Subedge *e, Subedge *f) const
{
  if (e == f)
    return false;
  if (e->head() == f->head())
    return e->forward;
  CVertex *v = e->tail();
  bool res;
  try {
    res = order(v, e, f);
  }
  catch (SignException &exc) {
    if (fail)
      throw signException;
    e->edge->edge->increasePrecision();
    f->edge->edge->increasePrecision();
    res = order(v, e, f);
    e->edge->edge->decreasePrecision();
    f->edge->edge->decreasePrecision();
  }
  return res;
}

bool SubedgeOrder::order (CVertex *v, Subedge *e, Subedge *f) const
{
  int coord = face->n.projectionCoordinate();
  if (v->iedges[1] ||
      v->iedges[0] && v->face == face) {
    int s1 = e->u().project(coord).getY().sign(),
      s2 = f->u().project(coord).getY().sign();
    if (s1 != s2)
      return s1 == 1;
  }
  else {
    if (e->edge->edge->bflag)
      return !e->forward;
    if (f->edge->edge->bflag)
      return f->forward;
  }
  return e->u().crossproduct(f->u(), coord).sign() == -1;
}

SFace::SFace (CFace *iface) : face(iface)
{
  hfaces[0].face = hfaces[1].face = this;
  hfaces[0].outer = true;
  hfaces[1].outer = false;
}

void SFace::addBoundary (Subedge *e)
{
  boundary.push_back(e);
  Subedge *f = e;
  do {
    f->face = this;
    f = f->twin->next;
  }
  while (f != e);
}

bool SFace::contains (CVertex *v, int coord, bool fail, Parameter &k) const
{
  bool res = false, kflag = true;
  for (Subedges::const_iterator b = boundary.begin(); b != boundary.end(); ++b) {
    Subedge *e = *b;
    CVertex *v0 = e->tail();
    if (v == v0)
      return false;
    Point2 t = (v->p - v0->p).project(coord);
    bool st;
    try {
      st = t.getY().sign() == -1;
    }
    catch (SignException &exc) {
      if (fail)
	throw signException;
      v->increasePrecision();
      v0->increasePrecision();
      Point2 t = (v->p - v0->p).project(coord);
      st = t.getY().sign() == -1;
      v->decreasePrecision();
      v0->decreasePrecision();
    }
    do {
      CVertex *w = e->head();
      if (v == w)
	return false;
      Point2 h = (v->p - w->p).project(coord);
      bool sh;
      try {
	sh = h.getY().sign() == -1;
      }
      catch (SignException &exc) {
	if (fail)
	  throw signException;
	v->increasePrecision();
	w->increasePrecision();
	Point2 h = (v->p - w->p).project(coord);
	sh = h.getY().sign() == -1;
	v->decreasePrecision();
	w->decreasePrecision();
      }
      if (st != sh) {
	Point2 u = e->edge->u().project(coord);
	int s3, s4;
	Parameter ke;
	try {
	  s3 = t.crossproduct(u).sign();
	  s4 = u.getY().sign();
	  if (s3 != s4)
	    ke = u.crossproduct(h)/u.getY();
	}
	catch (SignException &exc) {
	  if (fail)
	    throw signException;
	  CVertex *et = e->tail();
	  v->increasePrecision();
	  w->increasePrecision();
	  et->increasePrecision();
	  e->edge->increasePrecision();
	  Point2 t = (v->p - et->p).project(coord),
	    h = (v->p - w->p).project(coord), u = e->edge->u().project(coord);
	  s3 = t.crossproduct(u).sign();
	  s4 = u.getY().sign();
	  if (s3 != s4)
	    ke = u.crossproduct(h)/u.getY();
	  v->decreasePrecision();
	  w->decreasePrecision();
	  et->decreasePrecision();
	  e->edge->decreasePrecision();
	  ke.decreasePrecision();
	}
	if (s3 != s4) {
	  res = !res;
	  if (kflag) {
	    k = ke;
	    kflag = false;
	  }
	  else 
	    try {
	      if (ke < k)
		k = ke;
	    }
	    catch (SignException &exc) {
	    }
	}
      }
      t = h;
      st = sh;
      e = e->twin->next;
    }
    while (e != *b);
    if (!res)
      return false;
  }
  return true;
}

void SFace::triangulate (Triangles &tr) const
{
  vector<CVertices *> reg;
  for (Subedges::const_iterator b = boundary.begin(); b != boundary.end(); ++b) {
    Subedge *e = *b;
    CVertices *rb = new CVertices;
    reg.push_back(rb);
    do {
      rb->push_back(e->tail());
      e = e->twin->next;
    }
    while (e != *b);
  }
  if (reg.size() == 1 && reg[0]->size() == 3)
    tr.push_back(Triangle(reg[0]->at(0), reg[0]->at(1), reg[0]->at(2)));
  else {
    int coord = face->n.projectionCoordinate();
    ::triangulate(reg, coord, tr);
  }
  for (int i = 0; i < reg.size(); ++i)
    delete reg[i];
}

HSFace * HSFace::twin () const
{ 
  return outer ? face->hfaces + 1 : face->hfaces;
}

Shell * HSFace::formShell (bool allfaces)
{
  Shell *shell = new Shell;
  HSFaces st;
  st.push_back(this);
  bool manifold = true, outers = true;
  while (!st.empty()) {
    HSFace *f = *(st.end()-1);
    st.pop_back();
    if (f->flag)
      continue;
    f->flag = true;
    outers = outers && f->outer;
    shell->faces.push_back(f);
    bool fflag = f->neighbors(allfaces, st);
    manifold = manifold && fflag;
  }
  if (manifold && (allfaces || outers)) {
    shell->init();
    return shell;
  }
  delete shell;
  return 0;  
}

bool HSFace::neighbors (bool allfaces, HSFaces &res) const
{
  bool flag = true;
  for (Subedges::iterator b = face->boundary.begin(); 
       b != face->boundary.end(); ++b) {
    Subedge *e = *b;
    do {
      HSFace *f = neighbor(e, allfaces);
      if (!f)
	flag = false;
      else if (!f->flag)
	res.push_back(f);
      e = e->twin->next;
    }
    while (e != *b);
  }
  return flag;
}

HSFace * HSFace::neighbor (Subedge *e, bool allfaces) const
{
  return e->edge->edge->bflag ? neighborB(e, allfaces)
    : neighborI(e, allfaces);
}


HSFace * HSFace::neighborB (Subedge *e, bool allfaces) const
{
  int s = outer ? 1 : -1;
  CFace *a = face->face, *b = 0;
  CEdge *ee = e->edge, *fb = 0;
  MetaCEdge *eee = ee->edge;
  for (int i = 0; i < eee->ne; ++i) {
    CEdge *f = eee->getEdge(i);
    if (f->forward != ee->forward) {
      CFace *c = f->face;
      if (!b ||
	  (a->coplanar(b) || ee->ccw(b, a) == 1 ? 
	   ee->ccw(c, a) == s && ee->ccw(c, b) == s :
	   ee->ccw(c, a) == s || ee->ccw(c, b) == s)) {
	b = c;
	fb = f;
      }
    }
  }
  if (!allfaces && b)
    for (int i = 0; i < eee->ne; ++i) {
      CEdge *f = eee->getEdge(i);
      if (f != ee && f->forward == ee->forward) {
        CFace *c = f->face;
	if (a->coplanar(b) || ee->ccw(b, a) == 1 ?
	    ee->ccw(c, a) == - s && ee->ccw(c, b) == - s :
	    ee->ccw(c, a) == - s || ee->ccw(c, b) == - s)
	  return 0;
      }
    }
  if (!b)
    return 0;
  SFace *bs = fb->findSubedge(e->head(), e->tail())->face;
  if (!bs)
    return 0;
  return outer ? bs->hfaces : bs->hfaces + 1;
}

HSFace * HSFace::neighborI (Subedge *e, bool allfaces) const
{
  CVertex *t = e->tail(), *h = e->head();
  CEdge *ee = e->edge;
  MetaCEdge *eee = ee->edge;
  CEdge *f = ee == eee->edges[0] ? eee->edges[1] : eee->edges[0];
  if (!allfaces) {
    SFace *g = f->findSubedge(h, t)->face;
    if (!g)
      return 0;
    return outer ? g->hfaces : g->hfaces + 1;
  }
  Subedge *s = e->forward ? f->findSubedge(h, t) : f->findSubedge(t, h);
  SFace *g = outer ? s->face : s->twin->face;
  return e->forward ? g->hfaces : g->hfaces + 1;
}

bool HSFace::aboveZ (CVertex *v) const
{
  Parameter k;
  if (v->face == face->face || face->face->boundaryVertex(v) ||
      !face->contains(v, 3, false, k))
    return false;
  CVertex *a = face->boundary[0]->tail();
  int s1, s2;
  try {
    s1 = face->face->n.innerproduct(v->p - a->p).sign();
    s2 = face->face->n.z.sign();
  }
  catch (SignException &exc) {
    face->face->increasePrecision();
    a->increasePrecision();
    v->increasePrecision();
    s1 = face->face->n.innerproduct(v->p - a->p).sign();
    s2 = face->face->n.z.sign();
    face->face->decreasePrecision();
    a->decreasePrecision();
    v->decreasePrecision();
  }
  return s1 != s2;
}

void HSFace::triangulate (Triangles &tr) const
{
  int n = tr.size();
  face->triangulate(tr);
  if (!outer)
    for (int i = n; i < tr.size(); ++i) {
      CVertex *temp = tr[i].a;
      tr[i].a = tr[i].b;
      tr[i].b = temp;
    }
}

void Shell::init ()
{
  v = 0;
  for (HSFaces::iterator f = faces.begin(); f != faces.end(); ++f) {
    for (Subedges::iterator b = (*f)->face->boundary.begin();
	 b != (*f)->face->boundary.end(); ++b) {
      Subedge *e = *b;
      do {
	CVertex *t = e->tail();
	e = e->twin->next;
	if (!v || v->orderZ(t))
	  v = t;
      }
      while (e != *b);
    }
  }
  if (v->iedges[0]) {
    outer = true;
    return;
  }
  CVertex *w = 0;
  for (HSFaces::iterator f = faces.begin(); f != faces.end(); ++f) {
    for (Subedges::iterator b = (*f)->face->boundary.begin();
	 b != (*f)->face->boundary.end(); ++b) {
      Subedge *e = *b;
      do {
	CVertex *t = e->tail(), *h = e->head();
	e = e->twin->next;
	if (t == v && (!w || w != h && v->steeperZ(h, w)))
	  w = h;
	else if (h == v && (!w || w != t && v->steeperZ(t, w)))
	  w = t;
      }
      while (e != *b);
    }
  }
  HSFace *g1 = 0, *g2 = 0;
  for (HSFaces::iterator f = faces.begin(); f != faces.end(); ++f) {
    for (Subedges::iterator b = (*f)->face->boundary.begin();
	 b != (*f)->face->boundary.end(); ++b) {
      Subedge *e = *b;
      do {
	CVertex *t = (*f)->outer ? e->tail() : e->head(),
	  *h = (*f)->outer ? e->head() : e->tail();
	e = e->twin->next;
	if (t == v && h == w)
	  g1 = *f;
	else if (t == w && h == v)
	  g2 = *f;
      }
      while (e != *b);
    }
  }
  try {
    outer = isOuter(w, g1, g2);
  }
  catch (SignException &excp) {
    v->increasePrecision();
    w->increasePrecision();
    g1->face->face->increasePrecision();
    g2->face->face->increasePrecision();
    outer = isOuter(w, g1, g2);
    v->decreasePrecision();
    w->decreasePrecision();
    g1->face->face->decreasePrecision();
    g2->face->face->decreasePrecision();
  }
}

bool Shell::isOuter (CVertex *w, HSFace *g1, HSFace *g2)
{
  CFace *f1 = g1->face->face, *f2 = g2->face->face;
  Point3 n1 = g1->outer ? f1->n : - f1->n,
    n2 = g2->outer ? f2->n : - f2->n;
  if (f1->ffa && !f1->fv && f1->ffa->id == f2->ffa->id)
    return n1.z.sign() == -1;
  return (v->p - w->p).circulation(n1, n2).sign() == 1;
}

CVertex * Shell::boundaryVertex () const
{
  for (HSFaces::const_iterator f = faces.begin(); f != faces.end(); ++f)
    for (Subedges::iterator b = (*f)->face->boundary.begin();
	 b != (*f)->face->boundary.end(); ++b) {
      Subedge *e = *b;
      do {
	if (!e->tail()->face)
	  return e->tail();
	e = e->twin->next;
      }
      while (e != *b);
    }
  return 0;
}

Cell::~Cell ()
{ 
  for (Shells::iterator s = shells.begin(); s != shells.end(); ++s)
    delete *s;
}

void Cell::addShell (Shell *s)
{
  shells.push_back(s);
  for (HSFaces::iterator f = s->faces.begin(); f != s->faces.end(); ++f)
    (*f)->cell = this;
}

void Cell::setBBox ()
{
  bbox[0] = bbox[2] = bbox[4] = 1e10;
  bbox[1] = bbox[3] = bbox[5] = -1e10;
  HSFaces &faces = shells[0]->faces;
  for (HSFaces::iterator f = faces.begin(); f != faces.end(); ++f)
    for (Subedges::iterator b = (*f)->face->boundary.begin();
	 b != (*f)->face->boundary.end(); ++b) {
      Subedge *e = *b;
      do {
	CVertex *v = e->tail();
	e = e->twin->next;
	bbox[0] = min(bbox[0], v->p.x.lb());
	bbox[1] = max(bbox[1], v->p.x.ub());
	bbox[2] = min(bbox[2], v->p.y.lb());
	bbox[3] = max(bbox[3], v->p.y.ub());
	bbox[4] = min(bbox[4], v->p.z.lb());
	bbox[5] = max(bbox[5], v->p.z.ub());
      }
      while (e != *b);
    }
}

bool Cell::contains (CVertex *v) const
{
  if (!bboxContains(bbox, v->p))
    return false;
  for (HSFaces::const_iterator f = shells[0]->faces.begin(); 
	 f != shells[0]->faces.end(); ++f)
    for (Subedges::iterator b = (*f)->face->boundary.begin();
	 b != (*f)->face->boundary.end(); ++b) {
      Subedge *e = *b;
      do {
	if (e->tail() == v)
	  return false;
	e = e->twin->next;
      }
      while (e != *b);
    }
  bool res = false;
  for (HSFaces::const_iterator f = shells[0]->faces.begin(); 
	 f != shells[0]->faces.end(); ++f)
      if ((*f)->aboveZ(v))
	res = !res;
  return res;
}

void Cell::triangulate (Triangles &tr) const
{
  for (Shells::const_iterator s = shells.begin(); s != shells.end(); ++s)
    for (HSFaces::const_iterator f = (*s)->faces.begin(); 
	 f != (*s)->faces.end(); ++f)
      (*f)->triangulate(tr);
}


void formCells (const Shells &shells, Cells &cells)
{
  cells.push_back(new Cell);
  for (Shells::const_iterator s = shells.begin(); s != shells.end(); ++s)
    if ((*s)->outer) {
      Cell *c = new Cell;
      cells.push_back(c);
      c->addShell(*s);
      c->setBBox();
    }
  for (Shells::const_iterator s = shells.begin(); s != shells.end(); ++s)
    if (!(*s)->outer) {
      Cell *c = enclosingCell((*s)->v, cells);
      c->addShell(*s);
    }
}

Polyhedron * polyhedron (const Triangles &tr)
{
  CVPVMap vmap;
  Polyhedron *p = new Polyhedron;
  for (Triangles::const_iterator t = tr.begin(); t != tr.end(); ++t) {
    PVertex *a = getVertex(t->a, vmap, p), *b = getVertex(t->b, vmap, p),
      *c = getVertex(t->c, vmap, p);
    p->addFace(a, b, c);
  }
  return p;
}

PVertex * getVertex (CVertex *v, CVPVMap &vmap, Polyhedron *p)
{
  CVPVMap::iterator iter = vmap.find(v);
  if (iter != vmap.end())
    return iter->second;
  PVertex *w = p->addVertex(v->p);
  w->parent = v;
  vmap.insert(pair<CVertex *, PVertex *>(v, w));
  return w;
}

Polyhedron * repair (Polyhedron *p, Convolution &con)
{
  con.merge(p, true, true);
  con.subdivide(true, false);
  Shells shells;
  con.formShells(true, shells);
  Cells cells;
  formCells(shells, cells);
  labelCells(cells);
  Polyhedron *q = repairPoly(con);
  checkVertices(q);
  for (Cells::iterator c = cells.begin(); c != cells.end(); ++c)
    delete *c;
  return q;
}

Polyhedron * repairPoly (const Convolution &con)
{
  CVPVMap vmap;
  Polyhedron *p = repairPolyInit(con, vmap);
  for (CFaces::const_iterator f = con.faces.begin(); f != con.faces.end(); ++f) 
    for (int i = 0; i < (*f)->ns; ++i) {
      SFace *g = (*f)->sfaces[i];
      bool gin = g->hfaces[0].cell->ain > 0, gtin = g->hfaces[1].cell->ain > 0;
      if (gin && !gtin)
	repairPolyFace(p, vmap, g->hfaces[1]);
      else if (!gin && gtin)
	repairPolyFace(p, vmap, g->hfaces[0]);
    }
  return p;
}

Polyhedron * repairPolyInit (const Convolution &con, CVPVMap &vmap)
{
  Polyhedron *p = new Polyhedron;
  for (MetaCEdges::const_iterator e = con.edges.begin(); 
       e != con.edges.end(); ++e) {
    int id1 = (*e)->edges[0]->face->id, id2 = (*e)->edges[1]->face->id;
    if ((*e)->bflag) {
      PVertex *t = getVertex((*e)->tail, vmap, p),
	*h = getVertex((*e)->head, vmap, p);
      t->parent = h->parent = 0;
      t->fids.insert(id1);
      t->fids.insert(id2);
      h->fids.insert(id1);
      h->fids.insert(id2);
    }
    for (CVertices::iterator v = (*e)->vertices.begin(); 
	 v != (*e)->vertices.end(); ++v) {
      PVertex *w = getVertex(*v, vmap, p);
      w->fids.insert(id1);
      w->fids.insert(id2);
      if ((*e)->bflag)
	w->fids.insert((*v)->face->id);
    }
  }
  return p;
}

void repairPolyFace (Polyhedron *p, CVPVMap &vmap, const HSFace &f)
{
  Triangles tr;
  f.triangulate(tr);
  int id = f.face->face->id;
  for (Triangles::const_iterator t = tr.begin(); t != tr.end(); ++t) 
    p->addFace(getVertex(t->a, vmap, p),
	       getVertex(t->b, vmap, p),
	       getVertex(t->c, vmap, p),
	       id);
}

void checkVertices (Polyhedron *p)
{
  int m = p->vertices.size();
  for (int i = 0; i < m; ++i)
    checkVertex(p, p->vertices[i]);
}

void checkVertex (Polyhedron *p, PVertex *v)
{
  for (PEdges::iterator e = v->edges.begin(); e != v->edges.end(); ++e)
    (*e)->face->flag = false;
  vector<PEdges> groups;
  for (PEdges::iterator e = v->edges.begin(); e != v->edges.end(); ++e)
    if (!(*e)->face->flag) {
      PEdges g;
      PEdge *f = *e;
      do {
        g.push_back(f);
        f->face->flag = true;
        f = f->face->boundary[(f->faceIndex() + 2)%3]->twin;
      }
      while (f != *e);
      if (g.size() == v->edges.size())
        return;
      groups.push_back(g);
    }
  v->edges.clear();
  for (int i = 0; i < groups.size(); ++i)
    updateCone(p, groups[i]);
}

void updateCone (Polyhedron *p, PEdges &edges)
{
  //cerr << "missing updateCone" << endl;
}

Polyhedron * minkowskiSum (Polyhedron *a, Polyhedron *b, double dmin, 
			   bool inner)
{
  Convolution acon, bcon, con;
  Polyhedron *ra = repair(a, acon), *rb = repair(b, bcon);
  con.sum(ra, rb);
  con.subdivide(false, false);
  Shells shells, mshells;
  con.formShells(false, shells);
  minkowskiShells(ra, rb, inner, shells, mshells);
  Cells cells;
  formCells(mshells, cells);
  Triangles tr;
  for (Cells::iterator c = cells.begin(); c != cells.end(); ++c) {
    (*c)->triangulate(tr);
    delete *c;
  }
  Polyhedron *o = polyhedron(tr);
  o->simplify(dmin);
  delete ra;
  delete rb;
  return o;
}

void minkowskiShells (Polyhedron *a, Polyhedron *b, bool inner,
		      const Shells &shells, Shells &res)
{
  if (shells.empty())
    return;
  res.push_back(shells[0]);
  for (int i = 1; i < shells.size(); ++i)
    if (inner && minkowskiShell(a, b, shells[i]))
      res.push_back(shells[i]);
    else
      delete shells[i];
}

bool minkowskiShell (Polyhedron *a, Polyhedron *b, Shell *shell)
{
  CVertex *t = shell->boundaryVertex();
  if (!t)
    return false;
  PVertex *at, *bt = t->vb;
  Polyhedron *c = negativeTranslate(a, t, at);
  bool res = !c->intersects(b, at, bt);
  delete c;
  return res;
}

Polyhedron * negativeTranslate (Polyhedron *p, CVertex *t, PVertex *&at)
{
  Polyhedron *q = new Polyhedron;
  PVPVMap vmap;
  for (PVertices::const_iterator v = p->vertices.begin(); 
       v != p->vertices.end(); ++v) {
    PVertex *w = new PVertexD(t, *v);
    q->addVertex(w);
    vmap.insert(pair<PVertex *, PVertex *>(*v, w));
    if (*v == t->va)
      at = w;
  }
  for (PFaces::const_iterator f = p->faces.begin(); f != p->faces.end(); ++f) {
    PVPVMap::iterator a = vmap.find((*f)->boundary[0]->tail),
      b = vmap.find((*f)->boundary[1]->tail), c = vmap.find((*f)->boundary[2]->tail);
    q->addFace(c->second, b->second, a->second);
  }
  return q;
}

Polyhedron * setOperation (Polyhedron *a, Polyhedron *b, SetOp op, double dmin)
{
  Convolution acon, bcon, con;
  Polyhedron *ra = repair(a, acon), 
    *rb = repair(b, bcon);
  con.merge(ra, true, false);
  con.merge(rb, false, false);
  con.subdivide(true, true);
  Shells shells;
  con.formShells(true, shells);
  Cells cells;
  formCells(shells, cells);
  labelCells(cells);
  Triangles tr;
  for (CFaces::iterator f = con.faces.begin(); f != con.faces.end(); ++f)
    for (int i = 0; i < (*f)->ns; ++i) {
      SFace *g = (*f)->sfaces[i];
      Cell *c = g->hfaces[0].cell, *d = g->hfaces[1].cell;
      bool in1 = inSet(c->ain, c->bin, op), in2 = inSet(d->ain, d->bin, op);
      if (in1 && !in2)
	g->hfaces[1].triangulate(tr);
      else if (!in1 && in2)
	g->hfaces[0].triangulate(tr);
    }
  for (Cells::iterator c = cells.begin(); c != cells.end(); ++c)
    delete *c;
  Polyhedron *o = polyhedron(tr);
  o->simplify(dmin);
  delete ra;
  delete rb;
  return o;
}

Cell * enclosingCell (CVertex *v, const Cells &cells)
{
  for (int i = cells.size() - 1; i > 0; --i)
    if (cells[i]->contains(v))
      return cells[i];
  return cells[0];
}

void labelCells (Cells &cells)
{
  Cells st;
  st.push_back(cells[0]);
  cells[0]->flag = false;
  for (int i = 1; i < cells.size(); ++i)
    cells[i]->flag = true;
  while (!st.empty()) {
    Cell *c = *(st.end()-1);
    st.pop_back();
    for (Shells::iterator s = c->shells.begin(); s != c->shells.end(); ++s)
      for (HSFaces::iterator f = (*s)->faces.begin(); 
	   f != (*s)->faces.end(); ++f) {
	HSFace *g = (*f)->twin();
	Cell *d = g->cell;
	if (d->flag) {
	  d->flag = false;
	  if (g->face->face->flag) {
	    d->ain = g->outer ? c->ain - 1 : c->ain + 1;
	    d->bin = c->bin;
	  }
	  else {
	    d->ain = c->ain;
	    d->bin = g->outer ? c->bin - 1 : c->bin + 1;
	  }
	  st.push_back(d);
	}
      }
  }
}

bool inSet (int a, int b, SetOp op)
{
  bool ain = a > 0, bin = b > 0;
  switch (op) {
  case Union:
    return ain || bin;
  case Intersection:
    return ain && bin;
  case Complement:
    return ain && !bin;
  }
}

double pack3 (Polyhedron *a, Polyhedron *b, Polyhedron *c, double dmin, 
	      Point3 *t)
{
  double d[3] = {a->maxBBoxDimension(), b->maxBBoxDimension(),
		 c->maxBBoxDimension()};
  sort(d, d + 3);
  double lb = d[2], ub = d[1] + d[2] + 0.1;
  cerr << setprecision(16) << "initial interval: [" << lb << ", " << ub << "]"
       << endl;
  while (ub - lb > dmin) {
    double m = 0.5*(lb + ub), t0 = getTime();
    bool flag = pack3(a, b, c, dmin, t, m);
    double t1 = getTime() - t0;
    cerr << (flag ? "yes " : "no ") << m << "; time " << t1 << endl;
    if (flag)
      ub = m;
    else
      lb = m;
  }
  return ub;
}

bool pack3 (Polyhedron *a, Polyhedron *b, Polyhedron *c, double dmin, 
	    Point3 *t, double bs)
{
  double u01[6], u02[6], u03[6], v012[6], v013[6], v023[6],
    box[6] = {0.0, bs, 0.0, bs, 0.0, bs};
  a->freeSpace(box, u01);
  b->freeSpace(box, u02);
  c->freeSpace(box, u03);
  Polyhedron *ma = a->negative(), *mb = b->negative(),
    *mc = c->negative(), *u12 = minkowskiSum(a, mb, dmin, true),
    *u13 = minkowskiSum(a, mc, dmin, true), 
    *u23 = minkowskiSum(b, mc, dmin, true);
  freeSpace(u01, u02, v012);
  freeSpace(u01, u03, v013);
  freeSpace(u02, u03, v023);
  Polyhedron *b012 = makeBox(v012), *b013 = makeBox(v013), 
    *b023 = makeBox(v023),
    *v12 = setOperation(b012, u12, Complement, dmin),
    *v13 = setOperation(b013, u13, Complement, dmin), 
    *v23 = setOperation(b023, u23, Complement, dmin),
    *s123 = minkowskiSum(v12, v23, dmin, true),
    *w13 = setOperation(v13, s123, Intersection, dmin);
  bool res = w13->faces.size() > 0;
  if (res) {
    Point3 t13 = w13->interiorPoint();
    double u03t[6], v01[6], v01u02[6], u02t[6], w01[6];
    translateBox(u03, - t13, u03t);
    intersectBoxes(u01, u03t, v01);
    freeSpace(v01, u02, v01u02);
    Polyhedron *bv01u02 = makeBox(v01u02),
      *v23mt = v23->negativeTranslate(t13),
      *temp = setOperation(v23mt, bv01u02, Intersection, dmin),
      *w12 = setOperation(v12, temp, Intersection, dmin);
    if (w12->faces.empty())
      res = false;
    else {
      Point3 t12 = w12->interiorPoint();
      translateBox(u02, - t12, u02t);
      intersectBoxes(v01, u02t, w01);
      t[0] = midpointBox(w01);
      t[1] = t[0] + t12;
      t[2] = t[0] + t13;
      delete bv01u02;
      delete v23mt;
      delete temp;
      delete w12;
    }
  }
  delete ma;
  delete mb;
  delete mc;
  delete u12;
  delete u13;
  delete u23;
  delete b012;
  delete b013;
  delete b023;
  delete v12;
  delete v13;
  delete v23;
  delete s123;
  delete w13;
  return res;
}

// a moves with respect to b
void freeSpace (double *a, double *b, double *res)
{
  for (int i = 0; i < 3; ++i) {
    res[2*i] = b[2*i] - a[2*i+1];
    res[2*i+1] = b[2*i+1] - a[2*i];
  }
}

void translateBox (double *a, const Point3 &t, double *res)
{
  double tx = t.getX().round(), ty = t.getY().round(), tz = t.getZ().round();
  res[0] = a[0] + tx;
  res[1] = a[1] + tx;
  res[2] = a[2] + ty;
  res[3] = a[3] + ty;
  res[4] = a[4] + tz;
  res[5] = a[5] + tz;
}

void intersectBoxes (double *a, double *b, double *res)
{
  for (int i = 0; i < 3; ++i) {
    res[2*i] = max(a[2*i], b[2*i]);
    res[2*i+1] = min(a[2*i+1], b[2*i+1]);
  }
}

Point3 midpointBox (double *a)
{
  return Point3(0.5*(a[0] + a[1]), 0.5*(a[2] + a[3]), 0.5*(a[4] + a[5]));
}

Polyhedron * makeBox (double *b)
{
  Polyhedron *p = new Polyhedron;
  p->addVertex(b[0], b[2], b[4]);
  p->addVertex(b[1], b[2], b[4]);
  p->addVertex(b[1], b[3], b[4]);
  p->addVertex(b[0], b[3], b[4]);
  p->addVertex(b[0], b[2], b[5]);
  p->addVertex(b[1], b[2], b[5]);
  p->addVertex(b[1], b[3], b[5]);
  p->addVertex(b[0], b[3], b[5]);
  p->addFace(p->vertices[0], p->vertices[2], p->vertices[1]);
  p->addFace(p->vertices[0], p->vertices[3], p->vertices[2]);
  p->addFace(p->vertices[0], p->vertices[1], p->vertices[5]);
  p->addFace(p->vertices[0], p->vertices[5], p->vertices[4]);
  p->addFace(p->vertices[1], p->vertices[2], p->vertices[6]);
  p->addFace(p->vertices[1], p->vertices[6], p->vertices[5]);
  p->addFace(p->vertices[2], p->vertices[3], p->vertices[7]);
  p->addFace(p->vertices[2], p->vertices[7], p->vertices[6]);
  p->addFace(p->vertices[3], p->vertices[0], p->vertices[4]);
  p->addFace(p->vertices[3], p->vertices[4], p->vertices[7]);
  p->addFace(p->vertices[4], p->vertices[5], p->vertices[6]);
  p->addFace(p->vertices[4], p->vertices[6], p->vertices[7]);
  return p;
}

void pack3output (Polyhedron *a, Polyhedron *b, Polyhedron *c,
		  Point3 *t, ostream &ostr)
{
  VIMap vimap;
  Points3 pts;
  vector<int> data;
  Polyhedron *p[3] = {a->translate(t[0]), b->translate(t[1]),
		      c->translate(t[2])};
  for (int i = 0; i < 3; ++i) {
    for (PFaces::const_iterator f = p[i]->faces.begin();
	 f != p[i]->faces.end(); ++f)
      for (int j = 0; j < 3; ++j)
	data.push_back(getPoint(vimap, pts, (*f)->boundary[j]->tail));
    delete p[i];
  }
  outputVTK(pts, data, ostr);
}

// cube with dimensions [- s, s] and depression in top with depth d
Polyhedron * makeWell (double s, double d)
{
  Polyhedron *p = new Polyhedron;
  p->addVertex(Point3(- s, - s, - s));
  p->addVertex(Point3(s, - s, - s));
  p->addVertex(Point3(s, s, - s));
  p->addVertex(Point3(- s, s, - s));
  p->addVertex(Point3(- s, - s, s));
  p->addVertex(Point3(s, - s, s));
  p->addVertex(Point3(s, s, s));
  p->addVertex(Point3(- s, s, s));
  p->addVertex(Point3(0, 0, s - d));
  p->addFace(p->vertices[0], p->vertices[2], p->vertices[1]);
  p->addFace(p->vertices[0], p->vertices[3], p->vertices[2]);
  p->addFace(p->vertices[0], p->vertices[1], p->vertices[5]);
  p->addFace(p->vertices[0], p->vertices[5], p->vertices[4]);
  p->addFace(p->vertices[1], p->vertices[2], p->vertices[6]);
  p->addFace(p->vertices[1], p->vertices[6], p->vertices[5]);
  p->addFace(p->vertices[2], p->vertices[3], p->vertices[7]);
  p->addFace(p->vertices[2], p->vertices[7], p->vertices[6]);
  p->addFace(p->vertices[3], p->vertices[0], p->vertices[4]);
  p->addFace(p->vertices[3], p->vertices[4], p->vertices[7]);
  p->addFace(p->vertices[4], p->vertices[5], p->vertices[8]);
  p->addFace(p->vertices[5], p->vertices[6], p->vertices[8]);
  p->addFace(p->vertices[6], p->vertices[7], p->vertices[8]);
  p->addFace(p->vertices[7], p->vertices[4], p->vertices[8]);
  return p;
}

// pyramid with square base with dimensions [-s, s] and height h
Polyhedron * pyramid (double s, double h)
{
  Polyhedron *p = new Polyhedron;
  p->addVertex(Point3(- s, - s, 0));
  p->addVertex(Point3(s, - s, 0));
  p->addVertex(Point3(s, s, 0));
  p->addVertex(Point3(- s, s, 0));
  p->addVertex(Point3(0, 0, h));
  p->addFace(p->vertices[0], p->vertices[2], p->vertices[1]);
  p->addFace(p->vertices[0], p->vertices[3], p->vertices[2]);
  p->addFace(p->vertices[0], p->vertices[1], p->vertices[4]);
  p->addFace(p->vertices[1], p->vertices[2], p->vertices[4]);
  p->addFace(p->vertices[2], p->vertices[3], p->vertices[4]);
  p->addFace(p->vertices[3], p->vertices[0], p->vertices[4]);
  return p;
}

// inverted pyramid
Polyhedron * ipyramid (double s, double h)
{
  Polyhedron *p = new Polyhedron;
  p->addVertex(Point3(- s, - s, 0));
  p->addVertex(Point3(s, - s, 0));
  p->addVertex(Point3(s, s, 0));
  p->addVertex(Point3(- s, s, 0));
  p->addVertex(Point3(0, 0, - h));
  p->addFace(p->vertices[0], p->vertices[1], p->vertices[2]);
  p->addFace(p->vertices[0], p->vertices[2], p->vertices[3]);
  p->addFace(p->vertices[1], p->vertices[0], p->vertices[4]);
  p->addFace(p->vertices[2], p->vertices[1], p->vertices[4]);
  p->addFace(p->vertices[3], p->vertices[2], p->vertices[4]);
  p->addFace(p->vertices[0], p->vertices[3], p->vertices[4]);
  return p;
}

// old version with unneeded special case for triangulate edges

/*
bool PEdge::colinear (const PVertex *v) const
{
  if (v == tail || v == head())
    return true;
  if (face->id != twin->face->id)
    return v->onFace(face) && v->onFace(twin->face);
  set<int> sth, sthv;
  fids(sth);
  set_intersection(sth.begin(), sth.end(), v->fids.begin(), v->fids.end(),
		   inserter(sthv, sthv.begin()));
  if (sthv.size() > 1)
    return true;
  if (sthv.empty())
    return false;
  CVertex *a = (CVertex *) tail->parent, *b = (CVertex *) head()->parent,
    *c = (CVertex *) v->parent;
  PEdge *e = a && a->iedges[0] && !a->iedges[1] ? a->iedges[0]->edge->edge : 0, 
    *f = b && b->iedges[0] && !b->iedges[1] ? b->iedges[0]->edge->edge : 0,
    *g = c && c->iedges[0] && !c->iedges[1] ? c->iedges[0]->edge->edge : 0;
  int aid = e ? a->face->ffa->id : -1, bid = f ? b->face->ffa->id : -1,
    cid = g ? c->face->ffa->id : -1;
  if (e && f && aid == bid && v->fids.find(aid) != v->fids.end())
    return colinearEE(e, f, v);
  if (e && g && aid == cid && head()->fids.find(aid) != head()->fids.end())
    return colinearEE(e, g, head());
  if (f && g && bid == cid && tail->fids.find(bid) != tail->fids.end())
    return colinearEE(f, g, tail);
  return false;
}

bool colinearEE (PEdge *e, PEdge *f, const PVertex *v)
{
  set<int> lines[3];
  int n = colinearEEinit(e, f, lines);
  if (n == 0)
    return false;
  for (int i = 1; i < n; ++i)
    if (onLine(v, lines[i]))
      return true;
  CEdge **vedges = v->parent ? ((CVertex *) v->parent)->iedges : 0;
  if (!vedges || !vedges[0] || vedges[1])
    return false;
  PEdge *h = vedges[0]->edge->edge;
  bool found1 = false, found2 = false;
  for (int i = 0; !found1 && i < n; ++i)
    found1 = onLine(h->tail, lines[i]);
  for (int i = 0; found1 && !found2 && i < n; ++i)
    found2 = onLine(h->head(), lines[i]);
  return found1 && found2;      
}

int colinearEEinit (PEdge *e, PEdge *f, set<int> *lines)
{
  int n = colinearEEinit1(e, f, lines);
  if (n > 0)
    return n;
  n = colinearEEinit1(e, f->twin, lines);
  if (n > 0)
    return n;
  n = colinearEEinit1(f, e, lines);
  if (n > 0)
    return n;
  return colinearEEinit1(f, e->twin, lines);
}

int colinearEEinit1 (PEdge *e, PEdge *f, set<int> *lines)
{
  int n = 0;
  set<int> se, set;
  e->fids(se);
  set_intersection(se.begin(), se.end(), f->tail->fids.begin(),
		   f->tail->fids.end(), inserter(set, set.begin()));
  if (set.size() > 1) {
    e->fids(lines[0]);
    ++n;
    if (f->tail != e->tail) {
      set_intersection(f->head()->fids.begin(), f->head()->fids.end(),
		       e->tail->fids.begin(), e->tail->fids.end(),
		       inserter(lines[n], lines[n].begin()));
      ++n;
    }
    if (f->head() != e->tail) {
      set_intersection(f->head()->fids.begin(), f->head()->fids.end(),
		       e->head()->fids.begin(), e->head()->fids.end(),
		       inserter(lines[n], lines[n].begin()));
      ++n;
    }
  }
  return n;
}

bool onLine (const PVertex *v, const set<int> &line)
{
  set<int> s;
  set_intersection(v->fids.begin(), v->fids.end(), line.begin(),
		   line.end(), inserter(s, s.begin()));
  return s.size() > 1;
}
*/

// debug



template<class T>
int find (T *x, vector<T *> &v)
{
  for (int i = 0; i < v.size(); ++i)
    if (v[i] == x)
      return i;
  return -1;
}

int findVertex (PVertex *x, PVertices &v) { return find<PVertex>(x, v); }

int findVertex (CVertex *x, CVertices &v) { return find<CVertex>(x, v); }

int findEdge (PEdge *x, PEdges &e) { return find<PEdge>(x, e); }

int findEdge (MetaCEdge *x, MetaCEdges &e) { return find<MetaCEdge>(x, e); }

int findEdge (CEdge *x, CEdges &e) { return find<CEdge>(x, e); }

int findFace (PFace *x, PFaces &f) { return find<PFace>(x, f); }

int findFace (CFace *x, CFaces &f) { return find<CFace>(x, f); }

int findFace (HSFace *x, HSFaces &f) { return find<HSFace>(x, f); }

int findShell (Shell *x, Shells &s) { return find<Shell>(x, s); }

int findCell (Cell *x, Cells &c) { return find<Cell>(x, c); }

void pvs (CVertices &v)
{
  cerr << "(";
  for (int i = 0; i < v.size(); ++i)
    pp(v[i]->p);
  cerr << ")" << endl;
}

void pe (PEdge *e)
{
  cerr << "(";
  pp(e->tail->p);
  pp(e->head()->p);
  cerr << ")";
}

void pes (const PEdges &ed)
{
  cerr << "(";
  for (int i = 0; i < ed.size(); ++i)
    pe(ed[i]);
  cerr << ")" << endl;
}

void pe (CEdge *e)
{
  cerr << "(";
  pp(e->tail()->p);
  pp(e->head()->p);
  cerr << ")";
}

void pes (const CEdges &ed)
{
  cerr << "(";
  for (int i = 0; i < ed.size(); ++i)
    pe(ed[i]);
  cerr << ")" << endl;
}

void pe (Subedge *e)
{
  cerr << "(";
  pp(e->tail()->p);
  pp(e->head()->p);
  cerr << ")";
}

void pes (const Subedges &ed)
{
  cerr << "(";
  for (int i = 0; i < ed.size(); ++i)
    pe(ed[i]);
  cerr << ")" << endl;
}

void pf (PFace *f)
{
  cerr << "(";
  for (int i = 0; i < 3; ++i)
    pe(f->boundary[i]);
  cerr << ")";
}

void pf (CFace *f)
{
  cerr << "(";
  for (int i = 0; i < f->nb; ++i)
    pe(f->boundary[i]);
  cerr << ")";
}

void pfs (CFaces &faces)
{
  cerr << "(";
  for (int i = 0; i < faces.size(); ++i)
    pf(faces[i]);
  cerr << ")" << endl;
}

void pf (SFace *f)
{
  cerr << "(";
  for (Subedges::iterator b = f->boundary.begin(); b != f->boundary.end(); ++b) {
    cerr << "(";
    Subedge *e = *b;
    do {
      cerr << e->tail()->p;
      e = e->twin->next;
    }
    while (e != *b);
    cerr << ")";
  }
  cerr << ")" << endl;
}

void ps (HSFaces *s)
{
  cerr << "(";
  for (int i = 0; i < s->size(); ++i)
    pf(s->at(i)->face);
  cerr << ")";
}

void pfids (PVertex *v)
{
  for (set<int>::iterator i = v->fids.begin(); i != v->fids.end(); ++i)
    cerr << *i << " ";
  cerr << endl;
}

PFaces * dfs (PFace *f)
{
  PFaces *res = new PFaces;
  PFaces st;
  st.push_back(f);
  while (!st.empty()) {
    PFace *g = *(st.end() - 1);
    st.pop_back();
    if (!g || g->flag)
      continue;
    g->flag = true;
    res->push_back(g);
    for (int i = 0; i < 3; ++i)
      st.push_back(g->boundary[i]->twin->face);
  }
  return res;
}

void components (Polyhedron *p, vector<PFaces *> &res)
{
  for (PFaces::iterator f = p->faces.begin(); f != p->faces.end(); ++f)
    (*f)->flag = false;
  for (PFaces::iterator f = p->faces.begin(); f != p->faces.end(); ++f)
    if (!(*f)->flag)
      res.push_back(dfs(*f));
}

int ne (Polyhedron *p)
{
  return p->vertices.size() - p->edges.size() + p->faces.size();
}

int ne (PFaces &faces)
{
  set<PVertex *> v;
  for (PFaces::iterator f = faces.begin(); f != faces.end(); ++f)
    for (int i = 0; i < 3; ++i)
      v.insert((*f)->boundary[i]->tail);
  return v.size() - faces.size()/2;
}

void nes (Polyhedron *p)
{
  vector<PFaces *> comp;
  components(p, comp);
  for (int i = 0; i < comp.size(); ++i) {
    cerr << comp[i]->size() << " " << ne(*comp[i]) << endl;
    delete comp[i];
  }
}

void ppoly (Polyhedron *p)
{
  p->output(VTK, cout);
}

int countSfaces (Convolution &c)
{
  int n = 0;
  for (int i = 0; i < c.faces.size(); ++i)
    n += c.faces[i]->ns;
  return n;
}

bool checkIn (Convolution &con)
{
  bool res = true;
  for (int i = 0; i < con.faces.size(); ++i)
    for (int j = 0; j < con.faces[i]->ns; ++j) {
      SFace *f = con.faces[i]->sfaces[j];
      if (f->hfaces[1].cell->ain != 1 + f->hfaces[0].cell->ain) {
	res = false;
	cerr << "bad " << i << " " << j << endl;
      }
    }
  return res;
}

void printCells (Cells &cells)
{
  for (int i = 0; i < cells.size(); ++i) {
    cerr << i << " " << cells[i]->ain << " ";
    for (int j = 0; j < cells[i]->shells.size(); ++j)
      cerr << cells[i]->shells[j]->faces.size() << " ";
    cerr << endl;
  }
}

int ne (Shell *s)
{
  set<CVertex *> sv;
  int ne = 0, nf = s->faces.size();
  for (HSFaces::iterator f = s->faces.begin(); f != s->faces.end(); ++f) 
    for (Subedges::iterator b = (*f)->face->boundary.begin(); 
	 b != (*f)->face->boundary.end(); ++b) {
      Subedge *e = *b;
      do {
	++ne;
	sv.insert(e->tail());
	e = e->twin->next;
      }
      while (e != *b);
    }
  int nv = sv.size();
  return nv - ne/2 + nf;
}

bool checkCoplanar (Polyhedron *p)
{
  bool res = true;
  Convolution con;
  con.merge(p, true, false);
  for (int i = 0; i < con.faces.size(); ++i) {
    CFace *f = con.faces[i];
    for (int j = 0; j < con.vertices.size(); ++j) {
      CVertex *v = con.vertices[j];
      if (v->va->edges.empty() || f->boundaryVertex(v))
	continue;
      if (f->coplanarVertex(v)) {
	f->increasePrecision();
	v->increasePrecision();
	try {
	  int s = f->n.innerproduct(v->p - f->boundary[0]->tail()->p).sign();
	  int nv = findVertex(v->va, p->vertices);
	  cerr << "false coplanar vertex " << nv << " on face " << i << endl;
	  res = false;
	}
	catch (SignException &exc) {}
        f->decreasePrecision();
	v->decreasePrecision();
      }
      else 
	try {
	  int s = f->n.innerproduct(v->p - f->boundary[0]->tail()->p).sign();
	}
	catch (SignException &exc) {
	  f->increasePrecision();
	  v->increasePrecision();
	  try {
	    int s = f->n.innerproduct(v->p - f->boundary[0]->tail()->p).sign();
	  }
	  catch (SignException &exc) {
	    int nv = findVertex(v->va, p->vertices);
	    cerr << "missing coplanar vertex " << nv << " on face " << i << endl;
	    res = false;
	  }
	  f->decreasePrecision();
	  v->decreasePrecision();
	}
    }
  }
  return res;
}

void findFid (int id, Convolution &con)
{
  for (int i = 0; i < con.faces.size(); ++i)
    if (con.faces[i]->ffa->id == id)
      cerr << i << " ";
  cerr << endl;
}

void findFid (int id, Polyhedron *p)
{
  for (int i = 0; i < p->faces.size(); ++i)
    if (p->faces[i]->id == id)
      cerr << i << " ";
  cerr << endl;
}

bool checkColinear (PEdge *e, PVertex *v)
{
  bool flag = false;
  try {
    Point3 u = v->p - e->tail->p, w = u.crossproduct(e->u());
    w.circulation(u, e->u()).sign();
  }
  catch (SignException &exc) {
    e->increasePrecision();
    v->increasePrecision();
    try {
      Point3 u = v->p - e->tail->p, w = u.crossproduct(e->u());
      w.circulation(u, e->u()).sign();
    }
    catch (SignException &exc) {
      flag = true;
    }
    e->decreasePrecision();
    v->decreasePrecision();
  }
  return flag;
}

bool checkColinear (Polyhedron *p)
{
  bool res = true;
  for (int i = 0; i < p->vertices.size(); ++i) {
    PVertex *v = p->vertices[i];
    if (v->edges.empty()) continue;
    for (int j = 0; j < p->edges.size(); ++j) {
      PEdge *e = p->edges[j];
      bool flag1 = e->colinear(v), flag2 = checkColinear(e, v);
      if (flag1 != flag2) {
	cerr << (flag1 ? "false" : "missing") << " colinear edge " 
	     << j << " vertex " << i << endl;
	res = false;
      }
    }
  }
  return res;
}

int edgeIndex (Subedge *g) 
{
  SFace *f = g->face;
  int i = 0;
  for (Subedges::iterator b = f->boundary.begin(); b != f->boundary.end(); ++b) {
    Subedge *e = *b;
    do {
      if (g == e)
	return i;
      ++i;
      e = e->twin->next;
    }
    while (e != *b);
  }
  return -1;
}  

void ptriangles (const Triangles &tr)
{
  VIMap vimap;
  Points3 pts;
  vector<int> data;
  for (Triangles::const_iterator t = tr.begin(); t != tr.end(); ++t) {
    data.push_back(getPoint(vimap, pts, t->a));
    data.push_back(getPoint(vimap, pts, t->b));
    data.push_back(getPoint(vimap, pts, t->c));
  }
  outputVTK(pts, data, cout);
}

void psfaces (vector<SFace *> &faces)
{
  Triangles tr;
  for (int i = 0; i < faces.size(); ++i)
    faces[i]->triangulate(tr);
  ptriangles(tr);
}

void pshell (Shell *s)
{
  Triangles tr;
  for (HSFaces::iterator f = s->faces.begin(); f != s->faces.end(); ++f)
    (*f)->triangulate(tr);
  ptriangles(tr);
}

void pcon (Convolution &con)
{
  Triangles tr;
  for (CFaces::iterator f = con.faces.begin(); f != con.faces.end(); ++f) {
    tr.push_back(Triangle((*f)->boundary[0]->tail(),
			  (*f)->boundary[1]->tail(),
			  (*f)->boundary[2]->tail()));
    if ((*f)->nb == 4)
      tr.push_back(Triangle((*f)->boundary[0]->tail(),
			    (*f)->boundary[2]->tail(),
			    (*f)->boundary[3]->tail()));
  }
  ptriangles(tr);
}

void pcons (Convolution &con)
{
  vector<SFace *> faces;
  for (CFaces::iterator f = con.faces.begin(); f != con.faces.end(); ++f)
    for (int i = 0; i < (*f)->ns; ++i)
      faces.push_back((*f)->sfaces[i]);
  psfaces(faces);
}
