#ifndef TRIANGULATE
#define TRIANGULATE

#include "sum.h"

enum TVertexType {Start, End, Left, Right, VMerge, VSplit};

class TEdge;

class TVertex {
 public:
  TVertex (CVertex *iv, int icoord)
    : v(iv), coord(icoord), p(v->p.project(coord)), edge(0) {}
  bool operator< (TVertex &w);
  TVertexType type ();
  void link (TVertex *w, bool twinflag);
  void insert (TEdge *e);
  void increasePrecision ();
  void decreasePrecision ();

  CVertex *v;
  int coord;
  Point2 p;
  TEdge *edge;
  bool flag;
};

typedef vector<TVertex *> TVertices;

class TVertexOrder {
 public:
  bool operator() (TVertex *v, TVertex *w) const {
    return v != w && *w < *v;
  }
};

typedef map<CVertex *, TVertex *> TVertexMap;

TVertex * getTVertex (CVertex *v, int coord, TVertexMap &vmap);

class TEdgeOrder {
 public:
  bool operator() (TEdge *a, TEdge *b) const;
};

typedef set<TEdge *, TEdgeOrder> TEdgeSet;

class TEdge {
 public:
  TEdge () : tail(0), helper(0), twin(0), next(0) {}
  TEdge (TVertex *itail, TEdge *itwin, bool iin, bool iflag)
    : tail(itail), helper(0), twin(itwin), next(0), in(iin), flag(iflag) {}
  TVertex * head() const { return twin->tail; }
  bool increasingY () const { return *tail < *head(); }
  bool left (TVertex *v);
  bool ccw (TEdge *a);
  bool ccwAux (TEdge *a) const;
  bool inArc (TEdge *a, TEdge *b);
  void monotoneLoop (TVertices &v);
  void increasePrecision ();
  void decreasePrecision ();

  TVertex *tail, *helper;
  TEdge *twin, *next;
  bool in, flag;
};

void triangulate (vector<CVertices *> &reg, int coord, Triangles &tr);

void triangulateInit (vector<CVertices *> &reg, int coord, TVertices &vertices);

void makeMonotone (TVertices &vertices);

void makeMonotone (TVertex *v, TEdgeSet &sweep);

TEdge * findLeft (TVertex *v, TEdgeSet &sweep);

void deleteVertices (TVertices &vertices);

void triangulateMonotone (TVertices &vertices, Triangles &tr);

void triangulateLoop (const TVertices &v, Triangles &tr);

void triangulateAdd (TVertex *v, TVertices &stack, Triangles &tr);

void triangulatePop (TVertex *v, TVertices &stack, Triangles &tr, 
		      bool done);

Triangle triangle (TVertex *a, TVertex *b, TVertex *c, bool flag);

#endif
